# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import abc
import copy
import math as pymath

import numpy
import scipy.constants as constants
import scipy.integrate as integrate
import scipy.interpolate as interpolate
from scipy.special import erf, j0

from .. import config, utilities
from ..exception import InputError
from ..experiment import Experiment
from ..math import NormGauss1d, NormLorentz1d, heaviside, solve_quartic
from .smaterials import Layer, LayerStack


def startdelta(start, delta, num):
    end = start + delta * (num - 1)
    return numpy.linspace(start, end, int(num))


class Model(object):
    """
    generic model class from which further models can be derived from
    """

    def __init__(self, experiment, **kwargs):
        """
        constructor for a generical simulation model.
        currently only the experiment class describing the diffraction geometry
        is stored in the base class

        Parameters
        ----------
        experiment : Experiment
            class describing the diffraction geometry, energy and wavelength of
            the model
        resolution_width :  float, optional
            defines the width of the resolution
        I0 :                float, optional
            the primary beam flux/intensity
        background :        float, optional
            the background added to the simulation
        energy :            float or str
            the experimental energy in eV
        resolution_type :   {'Gauss', 'Lorentz'}, optional
            type of resolution function, default: Gauss
        """
        local_fit_params = {'resolution_width': 'width of the resolution',
                            'I0': 'primary beam intensity',
                            'background': 'background intensity',
                            'energy': 'x-ray energy in eV'}
        if not hasattr(self, 'fit_paramnames'):
            self.fit_paramnames = []
        self.fit_paramnames += local_fit_params
        valid_kwargs = copy.copy(local_fit_params)
        valid_kwargs['resolution_type'] = 'resolution function typ'
        utilities.check_kwargs(kwargs, valid_kwargs,
                               self.__class__.__name__)

        if experiment:
            self.exp = experiment
        else:
            self.exp = Experiment([1, 0, 0], [0, 0, 1])
        self.resolution_width = kwargs.get('resolution_width', 0)
        self.resolution_type = kwargs.get('resolution_type', 'Gauss')
        self.I0 = kwargs.get('I0', 1)
        self.background = kwargs.get('background', 0)
        if 'energy' in kwargs:
            self.energy = kwargs['energy']

    @property
    def energy(self):
        return self.exp.energy

    @energy.setter
    def energy(self, en):
        self.exp.energy = en

    def convolute_resolution(self, x, y):
        """
        convolve simulation result with a resolution function

        Parameters
        ----------
        x :     array-like
            x-values of the simulation, units of x also decide about the unit
            of the resolution_width parameter
        y :     array-like
            y-values of the simulation

        Returns
        -------
        array-like
            convoluted y-data with same shape as y
        """
        if self.resolution_width == 0:
            return y
        else:
            dx = numpy.mean(numpy.gradient(x))
            nres = int(20 * numpy.abs(self.resolution_width / dx))
            xres = startdelta(-10*self.resolution_width, dx, nres + 1)
            # the following works only exactly for equally spaced data points
            if self.resolution_type == 'Gauss':
                fres = NormGauss1d
            else:
                fres = NormLorentz1d
            resf = fres(xres, numpy.mean(xres), self.resolution_width)
            resf /= numpy.sum(resf)  # proper normalization for discrete conv.
            # pad y to avoid edge effects
            interp = interpolate.InterpolatedUnivariateSpline(x, y, k=1)
            nextmin = numpy.ceil(nres/2.)
            nextpos = numpy.floor(nres/2.)
            xext = numpy.concatenate(
                (startdelta(x[0]-dx, -dx, nextmin)[-1::-1],
                 x,
                 startdelta(x[-1]+dx, dx, nextpos)))
            ypad = numpy.asarray(interp(xext))
            return numpy.convolve(ypad, resf, mode='valid')

    def scale_simulation(self, y):
        """
        scale simulation result with primary beam flux/intensity and add a
        background.

        Parameters
        ----------
        y :     array-like
            y-values of the simulation

        Returns
        -------
        array-like
            scaled y-values
        """
        return y * self.I0 + self.background


class LayerModel(Model, utilities.ABC):
    """
    generic model class from which further thin film models can be derived from
    """

    def __init__(self, *args, **kwargs):
        """
        constructor for a thin film model. The arguments consist of a
        LayerStack or individual Layer(s). Optional parameters are specified
        in the keyword arguments.

        Parameters
        ----------
        *args :         LayerStack or Layers
            either one LayerStack or several Layer objects can be given
        **kwargs :      dict
            optional parameters for the simulation. ones not listed below are
            forwarded to the superclass.
        experiment :    Experiment, optional
            class containing geometry and energy of the experiment.
        surface_hkl :   list or tuple, optional
            Miller indices of the surface (default: (0, 0, 1))
        """
        exp = kwargs.pop('experiment', None)
        super().__init__(exp, **kwargs)
        self.lstack_params = []
        self.lstack_structural_params = False
        self.xlabelstr = 'x (1)'
        if len(args) == 1:
            if isinstance(args[0], Layer):
                self.lstack = LayerStack('Stack for %s'
                                         % self.__class__.__name__, *args)
            else:
                self.lstack = args[0]
        else:
            self.lstack = LayerStack('Stack for %s' % self.__class__.__name__,
                                     *args)

    @abc.abstractmethod
    def simulate(self):
        """
        abstract method that every implementation of a LayerModel has to
        override.
        """
        pass

    def _create_return(self, x, E, ai=None, af=None, Ir=None,
                       rettype='intensity'):
        """
        function to create the return value of a simulation. by default only
        the diffracted intensity is returned. However, optionally also the
        incidence and exit angle as well as the reflected intensity can be
        returned.

        Parameters
        ----------
        x :         array-like
            independent coordinate value for the convolution with the
            resolution function
        E :         array-like
            electric field amplitude (complex)
        ai, af :    array-like, optional
            incidence and exit angle of the XRD beam (in radians)
        Ir :        array-like, optional
            reflected intensity
        rettype :   {'intensity', 'field', 'all'}, optional
            type of the return value. 'intensity' (default): returns the
            diffracted beam flux convoluted with the resolution function;
            'field': returns the electric field (complex) without convolution
            with the resolution function, 'all': returns the electric field,
            ai, af (both in degree), and the reflected intensity.

        Returns
        -------
        return value depends on value of rettype.
        """
        if rettype == 'intensity':
            ret = self.scale_simulation(
                self.convolute_resolution(x, numpy.abs(E)**2))
        elif rettype == 'field':
            ret = E
        elif rettype == 'all':
            ret = (E, numpy.degrees(ai), numpy.degrees(af), Ir)
        return ret

    def get_polarizations(self):
        """
        return list of polarizations which should be calculated
        """
        if self.polarization == 'both':
            return ('S', 'P')
        else:
            return (self.polarization,)

    def join_polarizations(self, Is, Ip):
        """
        method to calculate the total diffracted intensity from the intensities
        of S and P-polarization.
        """
        if self.polarization == 'both':
            ret = (Is + self.Cmono * Ip) / (1 + self.Cmono)
        else:
            if self.polarization == 'S':
                ret = Is
            else:
                ret = Ip
        return ret


class KinematicalModel(LayerModel):
    """
    Kinematical diffraction model for specular and off-specular qz-scans. The
    model calculates the kinematical contribution of one (hkl) Bragg peak,
    however considers the variation of the structure factor for different 'q'.
    The surface geometry is specified using the Experiment-object given to the
    constructor.
    """

    def __init__(self, *args, **kwargs):
        """
        constructor for a kinematic thin film model. The arguments consist of a
        LayerStack or individual Layer(s). Optional parameters are specified in
        the keyword arguments.

        Parameters
        ----------
        *args :     LayerStack or Layers
            either one LayerStack or several Layer objects can be given
        **kwargs :  dict
            optional parameters; also see LayerModel/Model.
        experiment : Experiment
            Experiment class containing geometry and energy of the experiment.
        """
        super().__init__(*args, **kwargs)
        self.lstack_params += ['thickness', ]
        self.lstack_structural_params = True
        self.xlabelstr = r'momentum transfer $Q_z$ ($\mathrm{\AA^{-1}}$)'
        # precalc optical properties
        self._init_en = 0
        self.init_chi0()

    def init_chi0(self):
        """
        calculates the needed optical parameters for the simulation. If any of
        the materials/layers is changing its properties this function needs to
        be called again before another correct simulation is made. (Changes of
        thickness does NOT require this!)
        """
        if self._init_en != self.energy:  # recalc properties if energy changed
            self.chi0 = numpy.asarray([l.material.chi0(en=self.energy)
                                       for l in self.lstack])
            self._init_en = self.energy

    def _prepare_kincalculation(self, qz, hkl):
        """
        prepare kinematic calculation by calculating some helper values
        """
        rel = constants.physical_constants['classical electron radius'][0]
        rel *= 1e10
        k = self.exp.k0

        # determine q-inplane
        t = self.exp._transform
        ql0 = t(self.lstack[0].material.Q(*hkl))
        qinp = numpy.sqrt(ql0[0]**2 + ql0[1]**2)

        # calculate needed angles
        qv = numpy.asarray([t.inverse((ql0[0], ql0[1], q)) for q in qz])
        Q = numpy.linalg.norm(qv, axis=1)
        theta = numpy.arcsin(Q / (2 * k))
        domega = numpy.arctan2(qinp, qz)
        alphai, alphaf = (theta + domega, theta - domega)
        # calculate structure factors
        f = numpy.empty((len(self.lstack), len(qz)), dtype=numpy.complex)
        fhkl = numpy.empty(len(self.lstack), dtype=numpy.complex)
        for i, l in enumerate(self.lstack):
            m = l.material
            fhkl[i] = m.StructureFactor(m.Q(*hkl), en=self.energy) /\
                m.lattice.UnitCellVolume()
            f[i, :] = m.StructureFactorForQ(qv, en0=self.energy) /\
                m.lattice.UnitCellVolume()

        E = numpy.zeros(len(qz), dtype=numpy.complex)
        return rel, alphai, alphaf, f, fhkl, E, t

    def _get_qz(self, qz, alphai, alphaf, chi0, absorption, refraction):
        k = self.exp.k0
        q = qz.astype(numpy.complex)
        if absorption and not refraction:
            q += 1j * k * numpy.imag(chi0) / \
                numpy.sin((alphai + alphaf) / 2)
        if refraction:
            q = k * (numpy.sqrt(numpy.sin(alphai)**2 + chi0) +
                     numpy.sqrt(numpy.sin(alphaf)**2 + chi0))
        return q

    def simulate(self, qz, hkl, absorption=False, refraction=False,
                 rettype='intensity'):
        """
        performs the actual kinematical diffraction calculation on the Qz
        positions specified considering the contribution from a single Bragg
        peak.

        Parameters
        ----------
        qz :        array-like
            simulation positions along qz
        hkl :       list or tuple
            Miller indices of the Bragg peak whos truncation rod should be
            calculated
        absorption : bool, optional
            flag to tell if absorption correction should be used
        refraction : bool, optional
            flag to tell if basic refraction correction should be performed. If
            refraction is True absorption correction is also included
            independent of the absorption flag.
        rettype :   {'intensity', 'field', 'all'}
            type of the return value. 'intensity' (default): returns the
            diffracted beam flux convoluted with the resolution function;
            'field': returns the electric field (complex) without convolution
            with the resolution function, 'all': returns the electric field,
            ai, af (both in degree), and the reflected intensity.

        Returns
        -------
        array-like
            return value depends on the setting of `rettype`, by default only
            the calculate intensity is returned
        """
        self.init_chi0()
        rel, ai, af, f, fhkl, E, t = self._prepare_kincalculation(qz, hkl)
        # calculate interface positions
        z = numpy.zeros(len(self.lstack))
        for i, l in enumerate(self.lstack[-1:0:-1]):
            z[-i-2] = z[-i-1] - l.thickness

        # perform kinematical calculation
        for i, l in enumerate(self.lstack):
            q = self._get_qz(qz, ai, af, self.chi0[i], absorption, refraction)
            q -= t(l.material.Q(*hkl))[-1]

            if l.thickness == numpy.inf:
                E += fhkl[i] * numpy.exp(-1j * z[i] * q) / (1j * q)
            else:
                E += - fhkl[i] * numpy.exp(-1j * q * z[i]) * \
                    (1 - numpy.exp(1j * q * l.thickness)) / (1j * q)

        wf = numpy.sqrt(heaviside(ai) * heaviside(af) * rel**2 /
                        (numpy.sin(ai) * numpy.sin(af))) * E
        return self._create_return(qz, wf, ai, af, rettype=rettype)


class KinematicalMultiBeamModel(KinematicalModel):
    """
    Kinematical diffraction model for specular and off-specular qz-scans. The
    model calculates the kinematical contribution of several Bragg peaks on
    the truncation rod and considers the variation of the structure factor.
    In order to use a analytical description for the kinematic diffraction
    signal all layer thicknesses are changed to a multiple of the respective
    lattice parameter along qz. Therefore this description only works for (001)
    surfaces.
    """

    def __init__(self, *args, **kwargs):
        """
        constructor for a kinematic thin film model. The arguments consist of a
        LayerStack or individual Layer(s). Optional parameters are specified in
        the keyword arguments.

        Parameters
        ----------
        *args :     LayerStack or Layers
            either one LayerStack or several Layer objects can be given
        **kwargs :  dict
            optional parameters. see also LayerModel/Model.
        experiment : Experiment
            Experiment class containing geometry and energy of the experiment.
        surface_hkl : list or tuple
            Miller indices of the surface (default: (0, 0, 1))
        """
        self.surface_hkl = kwargs.pop('surface_hkl', (0, 0, 1))
        super().__init__(*args, **kwargs)

    def simulate(self, qz, hkl, absorption=False, refraction=True,
                 rettype='intensity'):
        """
        performs the actual kinematical diffraction calculation on the Qz
        positions specified considering the contribution from a full
        truncation rod

        Parameters
        ----------
        qz :            array-like
            simulation positions along qz
        hkl :           list or tuple
            Miller indices of the Bragg peak whos truncation rod should be
            calculated
        absorption :    bool, optional
            flag to tell if absorption correction should be used
        refraction :    bool, optional,
            flag to tell if basic refraction correction should be performed. If
            refraction is True absorption correction is also included
            independent of the absorption flag.
        rettype :       {'intensity', 'field', 'all'}
            type of the return value. 'intensity' (default): returns the
            diffracted beam flux convoluted with the resolution function;
            'field': returns the electric field (complex) without convolution
            with the resolution function, 'all': returns the electric field,
            ai, af (both in degree), and the reflected intensity.

        Returns
        -------
        array-like
            return value depends on the setting of `rettype`, by default only
            the calculate intensity is returned
        """
        self.init_chi0()
        rel, ai, af, f, fhkl, E, t = self._prepare_kincalculation(qz, hkl)

        # calculate interface positions for integer unit-cell thickness
        z = numpy.zeros(len(self.lstack))
        for i, l in enumerate(self.lstack[-1:0:-1]):
            lat = l.material.lattice
            a3 = t(lat.GetPoint(*self.surface_hkl))[-1]
            n3 = l.thickness // a3
            z[-i-2] = z[-i-1] - a3 * n3
            if config.VERBOSITY >= config.INFO_LOW and \
                    numpy.abs(l.thickness/a3 - n3) > 0.01:
                print('XU.KinematicMultiBeamModel: %s thickness changed from'
                      ' %.2fA to %.2fA (%d UCs)' % (l.name, l.thickness,
                                                    a3 * n3, n3))

        # perform kinematical calculation
        for i, l in enumerate(self.lstack):
            q = self._get_qz(qz, ai, af, self.chi0[i], absorption, refraction)
            lat = l.material.lattice
            a3 = t(lat.GetPoint(*self.surface_hkl))[-1]

            if l.thickness == numpy.inf:
                E += f[i, :] * a3 * numpy.exp(-1j * z[i] * q) /\
                    (1 - numpy.exp(1j * q * a3))
            else:
                n3 = l.thickness // a3
                E += f[i, :] * a3 * numpy.exp(-1j * z[i] * q) * \
                    (1 - numpy.exp(1j * q * a3 * n3)) /\
                    (1 - numpy.exp(1j * q * a3))

        wf = numpy.sqrt(heaviside(ai) * heaviside(af) * rel**2 /
                        (numpy.sin(ai) * numpy.sin(af))) * E
        return self._create_return(qz, wf, ai, af, rettype=rettype)


class SimpleDynamicalCoplanarModel(KinematicalModel):
    """
    Dynamical diffraction model for specular and off-specular qz-scans.
    Calculation of the flux of reflected and diffracted waves for general
    asymmetric coplanar diffraction from an arbitrary pseudomorphic multilayer
    is performed by a simplified 2-beam theory (2 tiepoints, S and P
    polarizations)

    No restrictions are made for the surface orientation.

    The first layer in the model is always assumed to be the semiinfinite
    substrate indepentent of its given thickness

    Note:
        This model should not be used in real life scenarios since the made
        approximations severely fail for distances far from the reference
        position.
    """

    def __init__(self, *args, **kwargs):
        """
        constructor for a diffraction model. The arguments consist of a
        LayerStack or individual Layer(s). Optional parameters are specified
        in the keyword arguments.

        Parameters
        ----------
        *args :     LayerStack or Layers
            either one LayerStack or several Layer objects can be given
        **kwargs:   dict
            optional parameters for the simulation
        I0 :        float, optional
            the primary beam intensity
        background : float, optional
            the background added to the simulation
        resolution_width : float, optional
            the width of the resolution (deg)
        polarization: {'S', 'P', 'both'}
            polarization of the x-ray beam. If set to 'both' also Cmono, the
            polarization factor of the monochromator should be set
        Cmono :     float, optional
            polarization factor of the monochromator
        energy :    float or str
            the experimental energy in eV
        experiment : Experiment
            Experiment class containing geometry of the sample; surface
            orientation!
        """
        if not hasattr(self, 'fit_paramnames'):
            self.fit_paramnames = []
        self.fit_paramnames += ['Cmono', ]
        self.polarization = kwargs.pop('polarization', 'S')
        self.Cmono = kwargs.pop('Cmono', 1)
        super().__init__(*args, **kwargs)
        self.xlabelstr = 'incidence angle (deg)'
        self.hkl = None
        self.chih = None
        self.chimh = None

    def set_hkl(self, *hkl):
        """
        To speed up future calculations of the same Bragg peak optical
        parameters can be pre-calculated using this function.

        Parameters
        ----------
        hkl :       list or tuple
            Miller indices of the Bragg peak for the calculation
        """
        if hkl != (None, ):
            if len(hkl) < 3:
                hkl = hkl[0]
                if len(hkl) < 3:
                    raise InputError("need 3 Miller indices")
            newhkl = numpy.asarray(hkl)
        else:
            newhkl = self.hkl

        if self.energy != self._init_en or numpy.any(newhkl != self.hkl):
            self.hkl = newhkl
            self._init_en = self.energy

            # calculate chih
            self.chih = {'S': [], 'P': []}
            self.chimh = {'S': [], 'P': []}
            for l in self.lstack:
                q = l.material.Q(self.hkl)
                thetaB = numpy.arcsin(numpy.linalg.norm(q) / 2 / self.exp.k0)
                ch = l.material.chih(q, en=self.energy, polarization='S')
                self.chih['S'].append(-ch[0] + 1j*ch[1])
                self.chih['P'].append((-ch[0] + 1j*ch[1]) *
                                      numpy.abs(numpy.cos(2*thetaB)))
                if not getattr(l, 'inversion_sym', False):
                    ch = l.material.chih(-q, en=self.energy, polarization='S')
                self.chimh['S'].append(-ch[0] + 1j*ch[1])
                self.chimh['P'].append((-ch[0] + 1j*ch[1]) *
                                       numpy.abs(numpy.cos(2*thetaB)))

            for pol in ('S', 'P'):
                self.chih[pol] = numpy.asarray(self.chih[pol])
                self.chimh[pol] = numpy.asarray(self.chimh[pol])

    def _prepare_dyncalculation(self, geometry):
        """
        prepare dynamical calculation by calculating some helper values
        """
        t = self.exp._transform
        ql0 = t(self.lstack[0].material.Q(*self.hkl))
        hx = numpy.sqrt(ql0[0]**2 + ql0[1]**2)
        if geometry == 'lo_hi':
            hx = -hx

        # calculate vertical diffraction vector components and strain
        hz = numpy.zeros(len(self.lstack))
        for i, l in enumerate(self.lstack):
            hz[i] = t(l.material.Q(*self.hkl))[2]
        return t, hx, hz

    def simulate(self, alphai, hkl=None, geometry='hi_lo', idxref=1):
        """
        performs the actual diffraction calculation for the specified
        incidence angles.

        Parameters
        ----------
        alphai :    array-like
            vector of incidence angles (deg)
        hkl :       list or tuple, optional
            Miller indices of the diffraction vector (preferable use set_hkl
            method to speed up repeated calculations of the same peak!)
        geometry :  {'hi_lo', 'lo_hi'}, optional
            'hi_lo' for grazing exit (default) and 'lo_hi' for grazing
            incidence
        idxref :    int, optional
            index of the reference layer. In order to get accurate peak
            position of the film peak you want this to be the index of the film
            peak (default: 1). For the substrate use 0.

        Returns
        -------
        array-like
            vector of intensities of the diffracted signal
        """
        self.set_hkl(hkl)

        # return values
        Ih = {'S': numpy.zeros(len(alphai)), 'P': numpy.zeros(len(alphai))}

        t, hx, hz = self._prepare_dyncalculation(geometry)
        epsilon = (hz[idxref] - hz) / hz

        k = self.exp.k0
        thetaB = numpy.arcsin(numpy.sqrt(hx**2 + hz[idxref]**2) / 2 / k)
        # asymmetry angle
        asym = numpy.arctan2(hx, hz[idxref])
        gamma0 = numpy.sin(asym + thetaB)
        gammah = numpy.sin(asym - thetaB)

        # deviation of the incident beam from the kinematical maximum
        eta = numpy.radians(alphai) - thetaB - asym

        for pol in self.get_polarizations():
            x = numpy.zeros(len(alphai), dtype=numpy.complex)
            for i, l in enumerate(self.lstack):
                beta = (2 * eta * numpy.sin(2 * thetaB) +
                        self.chi0[i] * (1 - gammah / gamma0) -
                        2 * gammah * (gamma0 - gammah) * epsilon[i])
                y = beta / 2 / numpy.sqrt(self.chih[pol][i] *
                                          self.chimh[pol][i]) /\
                    numpy.sqrt(numpy.abs(gammah) / gamma0)
                c1 = -numpy.sqrt(self.chih[pol][i] / self.chih[pol][i] *
                                 gamma0 / numpy.abs(gammah)) *\
                    (y + numpy.sqrt(y**2 - 1))
                c2 = -numpy.sqrt(self.chih[pol][i] / self.chimh[pol][i] *
                                 gamma0 / numpy.abs(gammah)) *\
                    (y - numpy.sqrt(y**2 - 1))
                kz2mkz1 = k * numpy.sqrt(self.chih[pol][i] *
                                         self.chimh[pol][i] / gamma0 /
                                         numpy.abs(gammah)) *\
                    numpy.sqrt(y**2 - 1)
                if i == 0:  # substrate
                    pp = numpy.abs(gammah) / gamma0 * numpy.abs(c1)**2
                    m = pp < 1
                    x[m] = c1[m]
                    m = pp >= 1
                    x[m] = c2[m]
                else:  # layers
                    cphi = numpy.exp(1j * kz2mkz1 * l.thickness)
                    x = (c1 * c2 * (cphi - 1) + xs * (c1 - cphi * c2)) /\
                        (cphi * c1 - c2 + xs * (1 - cphi))
                xs = x
            Ih[pol] = numpy.abs(x)**2 * numpy.abs(gammah) / gamma0

        ret = self.join_polarizations(Ih['S'], Ih['P'])
        return self.scale_simulation(self.convolute_resolution(alphai, ret))


class DynamicalModel(SimpleDynamicalCoplanarModel):
    """
    Dynamical diffraction model for specular and off-specular qz-scans.
    Calculation of the flux of reflected and diffracted waves for general
    asymmetric coplanar diffraction from an arbitrary pseudomorphic multilayer
    is performed by a generalized 2-beam theory (4 tiepoints, S and P
    polarizations)

    The first layer in the model is always assumed to be the semiinfinite
    substrate indepentent of its given thickness
    """

    def simulate(self, alphai, hkl=None, geometry='hi_lo',
                 rettype='intensity'):
        """
        performs the actual diffraction calculation for the specified
        incidence angles and uses an analytic solution for the quartic
        dispersion equation

        Parameters
        ----------
        alphai :    array-like
            vector of incidence angles (deg)
        hkl :       list or tuple, optional
            Miller indices of the diffraction vector (preferable use set_hkl
            method to speed up repeated calculations of the same peak!)
        geometry :  {'hi_lo', 'lo_hi'}, optional
            'hi_lo' for grazing exit (default) and 'lo_hi' for grazing
            incidence
        rettype :   {'intensity', 'field', 'all'}, optional
            type of the return value. 'intensity' (default): returns the
            diffracted beam flux convoluted with the resolution function;
            'field': returns the electric field (complex) without convolution
            with the resolution function, 'all': returns the electric field,
            ai, af (both in degree), and the reflected intensity.

        Returns
        -------
        array-like
            vector of intensities of the diffracted signal, possibly changed
            return value due the rettype setting!
        """
        if len(self.get_polarizations()) > 1 and rettype != "intensity":
            raise ValueError('XU:DynamicalModel: return type (%s) not '
                             'supported with multiple polarizations!')
            rettype = 'intensity'
        self.set_hkl(hkl)

        # return values
        Ih = {'S': numpy.zeros(len(alphai)), 'P': numpy.zeros(len(alphai))}
        Ir = {'S': numpy.zeros(len(alphai)), 'P': numpy.zeros(len(alphai))}

        t, hx, hz = self._prepare_dyncalculation(geometry)

        k = self.exp.k0
        kc = k * numpy.sqrt(1 + self.chi0)
        ai = numpy.radians(alphai)
        Kix = k * numpy.cos(ai)
        Kiz = -k * numpy.sin(ai)
        Khz = numpy.sqrt(k**2 - (Kix + hx)**2)
        pp = Khz / k
        mask = numpy.logical_and(pp > 0, pp < 1)
        ah = numpy.zeros(len(ai))  # exit angles
        ah[mask] = numpy.arcsin(pp[mask])

        nal = len(ai)
        for pol in self.get_polarizations():
            if pol == 'S':
                CC = numpy.ones(nal)
            else:
                CC = abs(numpy.cos(ai+ah))
            pom = k**4 * self.chih['S'] * self.chimh['S']
            if config.VERBOSITY >= config.INFO_ALL:
                print('XU.DynamicalModel: calc. %s-polarization...' % (pol))

            M = numpy.zeros((nal, 4, 4), dtype=numpy.complex)
            for j in range(4):
                M[:, j, j] = numpy.ones(nal)

            for i, l in enumerate(self.lstack[-1::-1]):
                jL = len(self.lstack) - 1 - i
                A4 = numpy.ones(nal)
                A3 = 2 * hz[jL] * numpy.ones(nal)
                A2 = (Kix + hx)**2 + hz[jL]**2 + Kix**2 - 2 * kc[jL]**2
                A1 = 2 * hz[jL] * (Kix**2 - kc[jL]**2)
                A0 = (Kix**2 - kc[jL]**2) *\
                     ((Kix + hx)**2 + hz[jL]**2 - kc[jL]**2) - pom[jL] * CC**2
                X = solve_quartic(A4, A3, A2, A1, A0)
                X = numpy.asarray(X).T

                kz = numpy.zeros((nal, 4), dtype=numpy.complex)
                kz[:, :2] = X[numpy.imag(X) <= 0].reshape(nal, 2)
                kz[:, 2:] = X[numpy.imag(X) > 0].reshape(nal, 2)

                P = numpy.zeros((nal, 4, 4), dtype=numpy.complex)
                phi = numpy.zeros((nal, 4, 4), dtype=numpy.complex)
                c = ((Kix**2)[:, numpy.newaxis] + kz**2 - kc[jL]**2) / k**2 /\
                    self.chimh['S'][jL] / CC[:, numpy.newaxis]
                if jL > 0:
                    for j in range(4):
                        phi[:, j, j] = numpy.exp(1j * kz[:, j] * l.thickness)
                else:
                    phi = numpy.tile(numpy.identity(4), (nal, 1, 1))
                P[:, 0, :] = numpy.ones((nal, 4))
                P[:, 1, :] = c
                P[:, 2, :] = kz
                P[:, 3, :] = c * (kz + hz[jL])

                if i == 0:
                    R = numpy.copy(P)
                else:
                    temp = numpy.linalg.inv(Ps)
                    try:
                        R = numpy.matmul(temp, P)
                    except AttributeError:
                        R = numpy.einsum('...ij,...jk', temp, P)
                try:
                    M = numpy.matmul(numpy.matmul(M, R), phi)
                except AttributeError:
                    M = numpy.einsum('...ij,...jk',
                                     numpy.einsum('...ij,...jk', M, R), phi)
                Ps = numpy.copy(P)

            B = numpy.zeros((nal, 4, 4), dtype=numpy.complex)
            B[..., :2] = M[..., :2]
            B[:, 0, 2] = -numpy.ones(nal)
            B[:, 1, 3] = -numpy.ones(nal)
            B[:, 2, 2] = Kiz
            B[:, 3, 3] = -Khz
            C = numpy.zeros((nal, 4))
            C[:, 0] = numpy.ones(nal)
            C[:, 2] = Kiz

            E = numpy.einsum('...ij,...j', numpy.linalg.inv(B), C)
            Ir[pol] = numpy.abs(E[:, 2])**2  # reflected intensity
            Ih[pol] = numpy.abs(E[:, 3])**2 * numpy.abs(Khz / Kiz) * mask

        if len(self.get_polarizations()) > 1 and rettype == "intensity":
            ret = numpy.sqrt(self.join_polarizations(Ih['S'], Ih['P']))
        else:
            ret = E[:, 3] * numpy.sqrt(numpy.abs(Khz / Kiz) * mask)

        return self._create_return(alphai, ret, ai, ah, Ir, rettype=rettype)


class SpecularReflectivityModel(LayerModel):
    """
    model for specular reflectivity calculations
    """

    def __init__(self, *args, **kwargs):
        """
        constructor for a reflectivity model. The arguments consist of a
        LayerStack or individual Layer(s). Optional parameters are specified
        in the keyword arguments.

        Parameters
        ----------
        args :      LayerStack or Layers
            either one LayerStack or several Layer objects can be given
        kwargs:     dict
            optional parameters for the simulation; supported are:
        I0 :        float, optional
            the primary beam intensity
        background : float, optional
            the background added to the simulation
        sample_width : float, optional
            width of the sample along the beam
        beam_width : float, optional
            beam width in the same units as the sample width
        offset :    float, optional
            angular offset of the incidence angle (deg)
        resolution_width : float, optional
            width of the resolution (deg)
        energy :    float or str
            x-ray energy  in eV
        """
        if not hasattr(self, 'fit_paramnames'):
            self.fit_paramnames = []
        self.fit_paramnames += ['sample_width', 'beam_width', 'offset']
        self.sample_width = kwargs.pop('sample_width', numpy.inf)
        self.beam_width = kwargs.pop('beam_width', 0)
        self.offset = kwargs.pop('offset', 0)
        super().__init__(*args, **kwargs)
        self.lstack_params += ['thickness', 'roughness', 'density']
        self.xlabelstr = 'incidence angle (deg)'
        # precalc optical properties
        self._init_en = 0
        self.init_cd()

    def init_cd(self):
        """
        calculates the needed optical parameters for the simulation. If any of
        the materials/layers is changing its properties this function needs to
        be called again before another correct simulation is made. (Changes of
        thickness and roughness do NOT require this!)
        """
        if self._init_en != self.energy:
            self.cd = numpy.asarray([-l.material.chi0(en=self.energy)/2
                                     for l in self.lstack])
            self._init_en = self.energy

    def simulate(self, alphai):
        """
        performs the actual reflectivity calculation for the specified
        incidence angles

        Parameters
        ----------
        alphai :    array-like
            vector of incidence angles

        Returns
        -------
        array-like
            vector of intensities of the reflectivity signal
        """
        self.init_cd()
        ns, np = (len(self.lstack), len(alphai))
        lai = alphai - self.offset
        # get layer properties
        t = numpy.asarray([l.thickness for l in self.lstack])
        sig = numpy.asarray([l.roughness for l in self.lstack])
        rho = numpy.asarray([l.density/l.material.density
                             for l in self.lstack])
        cd = self.cd

        sai = numpy.sin(numpy.radians(lai))

        if self.beam_width > 0:
            shape = self.sample_width * sai / self.beam_width
            shape[shape > 1] = 1
        else:
            shape = numpy.ones(np)

        ETs = numpy.ones(np, dtype=numpy.complex)
        ERs = numpy.zeros(np, dtype=numpy.complex)
        ks = -self.exp.k0 * numpy.sqrt(sai**2 - 2 * cd[0] * rho[0])

        for i in range(ns):
            if i < ns-1:
                k = -self.exp.k0 * numpy.sqrt(sai**2 - 2 * cd[i+1] * rho[i+1])
                phi = numpy.exp(1j * k * t[i+1])
            else:
                k = -self.exp.k0 * sai
                phi = numpy.ones(np)
            r = (k - ks) / (k + ks) * numpy.exp(-2 * sig[i]**2 * k * ks)
            ET = phi * (ETs + r * ERs)
            ER = (r * ETs + ERs) / phi
            ETs = ET
            ERs = ER
            ks = k

        R = shape * abs(ER / ET)**2
        return self.scale_simulation(self.convolute_resolution(lai, R))

    def densityprofile(self, nz, plot=False):
        """
        calculates the electron density of the layerstack from the thickness
        and roughness of the individual layers

        Parameters
        ----------
        nz :    int
            number of values on which the profile should be calculated
        plot :  bool, optional
            flag to tell if a plot of the profile should be created

        Returns
        -------
        z :     array-like
            z-coordinates, z = 0 corresponds to the surface
        eprof : array-like
            electron profile
        """
        if plot:
            try:
                from matplotlib import pyplot as plt
            except ImportError:
                plot = False
                if config.VERBOSITY >= config.INFO_LOW:
                    print("XU.simpack: Warning: plot "
                          "functionality not available")

        rel = constants.physical_constants['classical electron radius'][0]
        rel *= 1e10
        nl = len(self.lstack)

        # get layer properties
        t = numpy.asarray([l.thickness for l in self.lstack])
        sig = numpy.asarray([l.roughness for l in self.lstack])
        rho = numpy.asarray([l.density/l.material.density
                             for l in self.lstack])
        delta = numpy.real(self.cd)

        totT = numpy.sum(t[1:])
        zmin = -totT - 10 * sig[0]
        zmax = 5 * sig[-1]

        z = numpy.linspace(zmin, zmax, nz)
        pre_factor = 2 * numpy.pi / self.exp.wavelength**2 / rel * 1e24

        # generate delta-rho values and interface positions
        zz = numpy.zeros(nl)
        dr = numpy.zeros(nl)
        dr[-1] = delta[-1] * rho[-1] * pre_factor
        for i in range(nl-1, 0, -1):
            zz[i-1] = zz[i] - t[i]
            dr[i-1] = delta[i-1] * rho[i-1] * pre_factor

        # calculate profile from contribution of all interfaces
        prof = numpy.zeros(nz)
        w = numpy.zeros((nl, nz))
        for i in range(nl):
            s = (1 + erf((z - zz[i]) / sig[i] / numpy.sqrt(2))) / 2
            mask = s < (1 - 1e-10)
            w[i, mask] = s[mask] / (1 - s[mask])
            mask = numpy.logical_not(mask)
            w[i, mask] = 1e10

        c = numpy.ones((nl, nz))
        for i in range(1, nl):
            c[i, :] = w[i-1, :] * c[i-1, :]
        c0 = w[-1, :] * c[-1, :]
        norm = numpy.sum(c, axis=0) + c0

        for i in range(nl):
            c[i] /= norm

        for j in range(nz):
            prof[j] = numpy.sum(dr * c[:, j])

        if plot:
            plt.figure('XU:density_profile', figsize=(5, 3))
            plt.plot(z, prof, 'k-', lw=2, label='electron density')
            plt.xlabel(r'z ($\mathrm{\AA}$)')
            plt.ylabel(r'electron density (e$^-$ cm$^{-3}$)')
            plt.tight_layout()

        return z, prof


class DynamicalReflectivityModel(SpecularReflectivityModel):
    """
    model for Dynamical Specular Reflectivity Simulations.
    It uses the transfer Matrix methods as given in chapter 3
    "Daillant, J., & Gibaud, A. (2008). X-ray and Neutron Reflectivity"
    """
    def __init__(self, *args, **kwargs):
        """
        constructor for a reflectivity model. The arguments consist of a
        LayerStack or individual Layer(s). Optional parameters are specified
        in the keyword arguments.

        Parameters
        ----------
        args :      LayerStack or Layers
            either one LayerStack or several Layer objects can be given
        kwargs:     dict
            optional parameters for the simulation; supported are:
        I0 :        float, optional
            the primary beam intensity
        background : float, optional
            the background added to the simulation
        sample_width : float, optional
            width of the sample along the beam
        beam_width : float, optional
            beam width in the same units as the sample width
        resolution_width : float, optional
            width of the resolution (deg)
        energy :    float or str
            x-ray energy  in eV
        polarization:   ['P', 'S']
            x-ray polarization
        """
        self.polarization = kwargs.pop('polarization', 'P')
        super().__init__(*args, **kwargs)
        self._init_en_opt = 0
        self._setOpticalConstants()

    def _setOpticalConstants(self):
        if self._init_en_opt != self.energy:
            self.n_indices = numpy.asarray(
                [l.material.idx_refraction(en=self.energy)
                 for l in self.lstack])
            # append n = 1 for vacuum
            self.n_indices = numpy.append(self.n_indices, 1)[::-1]
            self._init_en_opt = self.energy

    def _getTransferMatrices(self, alphai):
        """
        Calculation of Refraction and Translation Matrices per angle per layer.
        """
        # Set heights for each layer
        heights = numpy.asarray([l.thickness for l in self.lstack[1:]])
        heights = numpy.cumsum(heights)[::-1]
        heights = numpy.insert(heights, 0, 0.)  # first interface is at z=0

        # set K-vector in each layer
        kz_angles = -self.exp.k0 * numpy.sqrt(numpy.asarray(
            [n**2 - numpy.cos(numpy.radians(alphai))**2
             for n in self.n_indices]).T)

        # set Roughness for each layer
        roughness = numpy.asarray([l.roughness for l in self.lstack[1:]])[::-1]
        roughness = numpy.insert(roughness, 0, 0.)  # first interface is at z=0

        # Roughness is approximated by a Gaussian Statistics model modification
        # of the transfer matrix elements using Groce-Nevot factors (GNF).

        GNF_factor_P = numpy.asarray(
            [[numpy.exp(-(kz_next - kz)**2 * (rough**2) / 2)
              for (kz, kz_next, rough) in zip(kz_1angle, kz_1angle[1:],
                                              roughness[1:])]
             for kz_1angle in kz_angles])

        GNF_factor_M = numpy.asarray(
            [[numpy.exp(-(kz_next + kz) ** 2 * (rough ** 2) / 2)
              for (kz, kz_next, rough) in zip(kz_1angle, kz_1angle[1:],
                                              roughness[1:])]
             for kz_1angle in kz_angles])

        if self.polarization == 'S':
            p_factor_angles = numpy.asarray(
                [[(kz + kz_next) / (2 * kz)
                  for kz, kz_next in zip(kz_1angle, kz_1angle[1:])]
                 for kz_1angle in kz_angles])

            m_factor_angles = numpy.asarray(
                [[(kz - kz_next) / (2 * kz)
                  for kz, kz_next in zip(kz_1angle, kz_1angle[1:])]
                 for kz_1angle in kz_angles])
        else:
            p_factor_angles = numpy.asarray(
                [[(n_next**2*kz + n**2*kz_next) / (2*n_next**2*kz)
                  for (kz, kz_next, n, n_next) in zip(kz_1angle, kz_1angle[1:],
                                                      self.n_indices,
                                                      self.n_indices[1:])]
                 for kz_1angle in kz_angles])
            m_factor_angles = numpy.asarray(
                [[(n_next**2*kz - n**2*kz_next) / (2*n_next**2*kz)
                  for (kz, kz_next, n, n_next) in zip(kz_1angle, kz_1angle[1:],
                                                      self.n_indices,
                                                      self.n_indices[1:])]
                 for kz_1angle in kz_angles])

        # Translation Matrices dim = (angle, layer, 2, 2)
        T_matrices = numpy.asarray(
            [[([numpy.exp(-1.j*kz*height), 0], [0, numpy.exp(1.j*kz*height)])
              for kz, height in zip(kz_1angle, heights)]
             for kz_1angle in kz_angles])

        R_matrices = numpy.asarray(
            [[([p, m], [m, p]) for p, m in zip(P_fact, M_fact)]
             for (P_fact, M_fact) in zip(p_factor_angles, m_factor_angles)])

        for R_mat, GNF_P, GNF_M in zip(R_matrices, GNF_factor_P, GNF_factor_M):
            R_mat[0, 0] = R_mat[0, 0] * GNF_P
            R_mat[0, 1] = R_mat[0, 1] * GNF_M
            R_mat[1, 0] = R_mat[1, 0] * GNF_M
            R_mat[1, 1] = R_mat[1, 1] * GNF_P

        return T_matrices, R_matrices

    def simulate(self, alphai):
        """
        Simulates the Dynamical Reflectivity as a function of angle of
        incidence

        Parameters
        ----------
        alphai :    array-like
            vector of incidence angles

        Returns
        -------
        reflectivity:   array-like
            vector of intensities of the reflectivity signal
        transmitivity: array-like
            vector of intensities of the transmitted signal
        """
        self._setOpticalConstants()
        lai = alphai - self.offset
        # Get Refraction and Translation Matrices for each angle of incidence
        if lai[0] < 1.e-5:
            lai[0] = 1.e-5  # cutoff

        T_matrices, R_matrices = self._getTransferMatrices(lai)

        # Calculate the Transfer Matrix
        M_angles = numpy.zeros((lai.size, 2, 2), dtype=numpy.complex128)
        for (angle, R), T in zip(enumerate(R_matrices), T_matrices):
            pairwiseRT = [numpy.dot(t, r) for r, t in zip(R, T)]
            M = numpy.identity(2, dtype=numpy.complex128)
            for pair in pairwiseRT:
                M = numpy.dot(M, pair)
            M_angles[angle] = M

        # Reflectance and Transmittance
        R = numpy.array([numpy.abs((M[0, 1] / M[1, 1]))**2 for M in M_angles])
        T = numpy.array([numpy.abs((1. / M[1, 1]))**2 for M in M_angles])

        return R, T

    def scanEnergy(self, energies, angle):
        # TODO: this is quite inefficient, too many calls to internal functions
        # TODO: DO not return normalized refelctivity
        """
        Simulates the Dynamical Reflectivity as a function of photon energy at
        fixed angle.

        Parameters
        ----------
        energies: np.ndarray or list
            photon energies (in eV).
        angle : float
            fixed incidence angle

        Returns
        -------
        reflectivity:   array-like
            vector of intensities of the reflectivity signal
        transmitivity: array-like
            vector of intensities of the transmitted signal
        """
        R_energies, T_energies = numpy.array([]), numpy.array([])
        for energy in energies:
            self.energy = energy
            self._setOpticalConstants()
            T_matrices, R_matrices = self._getTransferMatrices([angle, 0])
            T_matrix = T_matrices[0]
            R_matrix = R_matrices[0]
            pairwiseRT = [numpy.dot(t, r) for r, t in zip(R_matrix, T_matrix)]
            M = numpy.identity(2, dtype=numpy.complex128)
            for pair in pairwiseRT:
                M = numpy.dot(M, pair)
            R = numpy.abs(M[0, 1] / M[1, 1]) ** 2
            T = numpy.abs(1. / M[1, 1]) ** 2
            R_energies = numpy.append(R_energies, R)
            T_energies = numpy.append(T_energies, T)
        return R_energies, T_energies


class ResonantReflectivityModel(SpecularReflectivityModel):
    """
    model for specular reflectivity calculations
    CURRENTLY UNDER DEVELOPEMENT! DO NOT USE!
    """
    def __init__(self, *args, **kwargs):
        """
        constructor for a reflectivity model. The arguments consist of a
        LayerStack or individual Layer(s). Optional parameters are specified
        in the keyword arguments.

        Parameters
        ----------
        args :      LayerStack or Layers
            either one LayerStack or several Layer objects can be given
        kwargs:     dict
            optional parameters for the simulation; supported are:
        I0 :        float, optional
            the primary beam intensity
        background : float, optional
            the background added to the simulation
        sample_width : float, optional
            width of the sample along the beam
        beam_width : float, optional
            beam width in the same units as the sample width
        resolution_width : float, optional
            width of the resolution (deg)
        energy :    float or str
            x-ray energy  in eV
        polarization:   ['P', 'S']
            x-ray polarization
        """
        self.polarization = kwargs.pop('polarization', 'S')
        super().__init__(*args, **kwargs)

    def simulate(self, alphai):
        """
        performs the actual reflectivity calculation for the specified
        incidence angles

        Parameters
        ----------
        alphai :    array-like
            vector of incidence angles

        Returns
        -------
        array-like
            vector of intensities of the reflectivity signal
        """
        self.init_cd()
        ns, np = (len(self.lstack), len(alphai))
        lai = alphai - self.offset

        # get layer properties
        t = numpy.asarray([l.thickness for l in self.lstack])
        sig = numpy.asarray([l.roughness for l in self.lstack])
        rho = numpy.asarray([l.density/l.material.density
                             for l in self.lstack])
        cd = self.cd
        qzvec = 4 * numpy.pi * numpy.sin(numpy.radians(lai)) /\
            utilities.en2lam(self.energy)
        qvec = numpy.array([[0., 0., qz] for qz in qzvec])
        chihP = numpy.array([[l.material.chih(q, en=self.energy,
                                              polarization=self.polarization)
                              for q in qvec]
                             for l in self.lstack])

        if self.polarization in ['S', 'P']:
            cd = cd + chihP
        else:
            cd = cd

        sai = numpy.sin(numpy.radians(lai))

        if self.beam_width > 0:
            shape = self.sample_width * sai / self.beam_width
            shape[shape > 1] = 1
        else:
            shape = numpy.ones(np)

        ETs = numpy.ones(np, dtype=numpy.complex)
        ERs = numpy.zeros(np, dtype=numpy.complex)
        ks = -self.exp.k0 * numpy.sqrt(sai**2 - 2 * cd[0] * rho[0])

        for i in range(ns):
            if i < ns-1:
                k = -self.exp.k0 * numpy.sqrt(sai**2 - 2 * cd[i+1] * rho[i+1])
                phi = numpy.exp(1j * k * t[i+1])
            else:
                k = -self.exp.k0 * sai
                phi = numpy.ones(np)
            r = (k - ks) / (k + ks) * numpy.exp(-2 * sig[i]**2 * k * ks)
            ET = phi * (ETs + r * ERs)
            ER = (r * ETs + ERs) / phi
            ETs = ET
            ERs = ER
            ks = k

        R = shape * abs(ER / ET)**2
        return self.scale_simulation(self.convolute_resolution(lai, R))


class DiffuseReflectivityModel(SpecularReflectivityModel):
    """
    model for diffuse reflectivity calculations

    The 'simulate' method calculates the diffuse reflectivity on the specular
    rod in coplanar geometry in analogy to the SpecularReflectivityModel.

    The 'simulate_map' method calculates the diffuse reflectivity for a 2D set
    of Q-positions. This method can also calculate the intensity for other
    geometries, like GISAXS with constant incidence angle or a quasi
    omega/2theta scan in GISAXS geometry.
    """

    def __init__(self, *args, **kwargs):
        """
        constructor for a reflectivity model. The arguments consist of a
        LayerStack or individual Layer(s). Optional parameters are specified
        in the keyword arguments.

        Parameters
        ----------
        args :     LayerStack or Layers
            either one LayerStack or several Layer objects can be given
        kwargs :   dict
            optional parameters for the simulation; supported are:
        I0 :        float, optional
            the primary beam intensity
        background : float, optional
            the background added to the simulation
        sample_width : float, optional
            width of the sample along the beam
        beam_width : float, optional
            beam width in the same units as the sample width
        resolution_width : float, optional
            defines the width of the resolution (deg)
        energy :    float, optional
            sets the experimental energy (eV)
        H :         float, optional
            Hurst factor defining the fractal dimension of the roughness (0..1,
            very slow for H != 1 or H != 0.5), default: 1
        vert_correl : float, optional
            vertical correlation length in (Angstrom), 0 means full replication
        vert_nu :   float, optional
            exponent in the vertical correlation function
        method :    int, optional
            1..simple DWBA (default), 2..full DWBA (slower)
        vert_int :  int, optional
            0..no integration over the vertical divergence, 1..with integration
            over the vertical divergence
        qL_zero :   float, optional
            value of inplane q-coordinate which can be considered 0, using
            method 2 it is important to avoid exact 0 and this value will be
            used instead
        """
        if not hasattr(self, 'fit_paramnames'):
            self.fit_paramnames = []
        self.fit_paramnames += ['H', 'vert_correl', 'vert_nu']
        self.H = kwargs.pop('H', 1)
        self.vert_correl = kwargs.pop('vert_correl', 0)
        self.vert_nu = kwargs.pop('vert_nu', 0)
        self.method = kwargs.pop('method', 1)
        self.vert = kwargs.pop('vert_int', 0)
        self.qL_zero = kwargs.pop('qL_zero', 5e-5)
        super().__init__(*args, **kwargs)
        self.lstack_params += ['lat_correl', ]

    def _get_layer_prop(self):
        """
        helper function to obtain layer properties needed for all types of
        simulations
        """
        nl = len(self.lstack)
        self.init_cd()

        t = numpy.asarray([float(l.thickness) for l in self.lstack[nl:0:-1]])
        sig = [float(l.roughness) for l in self.lstack[nl::-1]]
        rho = [l.density/l.material.density for l in self.lstack[nl::-1]]
        delta = self.cd * numpy.asarray(rho)
        xiL = [float(l.lat_correl) for l in self.lstack[nl::-1]]

        return t, sig, rho, delta, xiL

    def simulate(self, alphai):
        """
        performs the actual diffuse reflectivity calculation for the specified
        incidence angles. This method always uses the coplanar geometry
        independent of the one set during the initialization.

        Parameters
        ----------
        alphai :    array-like
            vector of incidence angles

        Returns
        -------
        array-like
            vector of intensities of the reflectivity signal
        """
        lai = alphai - self.offset
        # get layer properties
        t, sig, rho, delta, xiL = self._get_layer_prop()

        deltaA = numpy.sum(delta[:-1]*t)/numpy.sum(t)
        lam = utilities.en2lam(self.energy)
        if self.method == 2:
            qL = [-abs(self.qL_zero), abs(self.qL_zero)]
        else:
            qL = [0, ]
        qz = 4 * numpy.pi / lam * numpy.sin(numpy.radians(lai))
        R = self._xrrdiffv2(lam, delta, t, sig, xiL, self.H, self.vert_correl,
                            self.vert_nu, None, qL, qz, self.sample_width,
                            self.beam_width, 1e-4, 1000, deltaA, self.method,
                            1, self.vert)
        R = R.mean(axis=0)
        return self.scale_simulation(self.convolute_resolution(lai, R))

    def simulate_map(self, qL, qz):
        """
        performs diffuse reflectivity calculation for the rectangular grid of
        reciprocal space positions define by qL and qz.  This method uses the
        method and geometry set during the initialization of the class.

        Parameters
        ----------
        qL :    array-like
            lateral coordinate in reciprocal space (vector with NqL components)
        qz :    array-like
            vertical coordinate in reciprocal space (vector with Nqz
            components)

        Returns
        -------
        array-like
            matrix of intensities of the reflectivity signal, with shape
            (len(qL), len(qz))
        """
        # get layer properties
        t, sig, rho, delta, xiL = self._get_layer_prop()

        deltaA = numpy.sum(delta[:-1]*t)/numpy.sum(t)
        lam = utilities.en2lam(self.energy)
        localqL = numpy.copy(qL)
        if self.method == 2:
            localqL[qL == 0] = self.qL_zero

        R = self._xrrdiffv2(lam, delta, t, sig, xiL, self.H, self.vert_correl,
                            self.vert_nu, None, localqL, qz, self.sample_width,
                            self.beam_width, 1e-4, 1000, deltaA, self.method,
                            1, self.vert)
        return self.scale_simulation(R)

    def _xrrdiffv2(self, lam, delta, thick, sigma, xiL, H, xiV, nu, alphai, qL,
                   qz, samplewidth, beamwidth, eps, nmax, deltaA, method, scan,
                   vert):
        """
        simulation of diffuse reflectivity from a rough multilayer. Exact or
        simplified DWBA, fractal roughness model, Ming model of the vertical
        correlation, various scattering geometries

        The used incidence and exit angles are stored in _smap_alphai,
        _smap_alphaf

        Parameters
        ----------
        lam :       float
            x-ray wavelength in Angstrom
        delta :     list or array-like
            vector with the 1-n values (N+1 components, 1st component..layer at
            the free surface, last component..substrate)
        thick :     list or array-like
            vector with thicknesses (N components)
        sigma :     list or array-like
            vector with rms roughnesses (N+1 components)
        xiL :       list or array-like
            vector with lateral correlation lengths (N+1 components)
        H :         float
            Hurst factor (scalar)
        xiV :       float
            vertical correlation: 0..full replication, > 0 and nu > 0.. see
            below, > 0 and nu = 0..vertical correlation length
        nu :        float
            exponent in the vertical correlation function
            exp(-abs(z_m-z_n)*(qL/max(qL))**nu/xiV)
        alphai :    float
            incidence angle (scalar for scan=2, ignored for scan=1, 3)
        qL :        array-like
            lateral coordinate in reciprocal space (vector with NqL components)
        qz :        array-like
            vertical coordinate in reciprocal space (vector with Nqz
            components)
        samplewidth : float
            width of the irradiated sample area (scalar), =0..the irradiated
            are is assumed constant
        beamwidth : float
            width of the primary beam
        eps :       float
            small number
        nmax :      int
            max number of terms in the Taylor series of the lateral correlation
            function
        deltaA :    complex
            effective value of 1-n in simple DWBA (ignored for method=2)
        method :    int
            1..simple DWBA, 2..full DWBA
        scan :      int
            1..standard coplanar geometry, 2..standard GISAXS geometry with
            constant incidence angle, =3..quasi omega/2theta scan in GISAXS
            geometry (incidence and central-exit angles are equal)
        vert :      int
            0..no integration over the vertical divergence, 1..with integration
            over the vertical divergence

        Returns
        -------
        diffint :   array-like
            diffuse reflectivity intensity matrix
        """
        # worker function definitions
        def coherent(alphai, K, delta, thick, N, NqL, Nqz):
            """
            calculate coherent reflection/transmission signal of a multilayer

            Parameters
            ----------
            alphai :    array-like
                matrix of incidence angles in radians (NqL x Nqz components)
            K :         float
                x-ray wave-vector (2*pi/lambda)
            delta :     array-like
                vector with the 1-n values (N+1 components, 1st
                component..layer at the free surface, last
                component..substrate)
            thick :     array-like
                vector with thicknesses (N components)
            N :         int
                number layers in the stack
            NqL :       int
                number of lateral q-points to calculate
            Nqz :       int
                number of vertical q-points to calculate

            Returns
            -------
            T, R, R0, k0, kz : array-like
                transmission, reflection, surface reflection, z-component of
                k-vector, z-component of k-vector in the material.
            """
            k0 = -K * numpy.sin(alphai)
            kz = numpy.zeros((N+1, NqL, Nqz), dtype=numpy.complex)
            T = numpy.zeros((N+1, NqL, Nqz), dtype=numpy.complex)
            R = numpy.zeros((N+1, NqL, Nqz), dtype=numpy.complex)
            for jn in range(N+1):
                kz[jn, ...] = -K * numpy.sqrt(numpy.sin(alphai)**2 -
                                              2 * delta[jn])

            T[N, ...] = numpy.ones((NqL, Nqz), dtype=numpy.complex)
            kzs = kz[N, ...]  # kz in substrate
            for jn in range(N-1, -1, -1):
                kzn = kz[jn, ...]
                tF = 2 * kzn / (kzn + kzs)
                rF = (kzn - kzs) / (kzn + kzs)
                phi = numpy.exp(1j * kzn * thick[jn])
                T[jn, ...] = phi / tF * (T[jn+1, ...] + rF * R[jn+1, ...])
                R[jn, ...] = 1 / phi / tF * (rF * T[jn+1, ...] + R[jn+1, ...])
                kzs = numpy.copy(kzn)

            tF = 2 * k0 / (k0 + kzn)
            rF = (k0 - kzn) / (k0 + kzn)
            T0 = 1 / tF * (T[0, ...] + rF * R[0, ...])
            R0 = 1 / tF * (rF * T[0, ...] + R[0, ...])

            T /= T0
            R /= T0
            R0 /= T0

            return T, R, R0, k0, kz

        def correl(a, b, L, H, eps, nmax, vert, K, NqL, Nqz, isurf):
            """
            correlation function

            Parameters
            ----------
            a :     array-like
                lateral correlation parameter
            b :     array-like
                vertical correlation parameter
            L :     float
                lateral correlation length
            H :     float
                Hurst factor (scalar)
            eps :   float
                small number (decides integration cut-off), typical 1e-3
            nmax :  int
                max number of terms in the Taylor series of the lateral
                correlation function
            vert :  int
                flag to tell decide if integration over vertical divergence is
                used: 0..no integration, 1..with integration
            K :     float
                length of the x-ray wave-vector (2*pi/lambda)
            NqL :   int
                number of lateral q-points to calculate
            Nqz :   int
                number of vertical q-points to calculate
            isurf : array-like
                array with NqL, Nqz flags to tell if there is a positive
                incidence and exit angle

            Returns
            -------
            psi :   array-like
                correlation function
            """
            psi = numpy.zeros((NqL, Nqz), dtype=numpy.complex)
            if H == 0.5 or H == 1:
                dpsi = numpy.zeros_like(psi, dtype=numpy.complex)
                m = isurf > 0
                n = 1
                s = numpy.copy(b)
                errm = numpy.inf
                if H == 1 and vert == 0:
                    def f(a, n):
                        return numpy.exp(-a**2/4/n) / 2 / n**2
                elif H == 0.5 and vert == 0:
                    def f(a, n):
                        return 1. / (n**2 + a**2)**(3/2.)
                elif H == 1 and vert == 1:
                    def f(a, n):
                        return numpy.sqrt(numpy.pi/n**3) * numpy.exp(-a**2/4/n)
                elif H == 0.5 and vert == 1:
                    def f(a, n):
                        return 2. / (n**2 + a**2)

                while errm > eps and n < nmax:
                    dpsi[m] = s[m] * f(a[m], n)
                    if n > 1:
                        errm = abs(numpy.max(dpsi[m] / psi[m]))
                    psi[m] += dpsi[m]
                    s[m] *= b[m]/float(n)
                    n += 1
            else:
                if vert == 0:
                    kern = kernel
                else:
                    kern = kernelvert
                for jL in range(NqL):
                    for jz in range(Nqz):
                        if isurf[jL, jz] == 1:
                            xmax = (-numpy.log(eps / b[jL, jz]))**(1/(2*H))
                            psi[jL, jz] = cquad(kern, 0.0, numpy.real(xmax),
                                                epsrel=eps, epsabs=0,
                                                limit=nmax, args=(a[jL, jz],
                                                b[jL, jz], H))

            if vert == 0:
                psi *= 2 * numpy.pi * L ** 2
            else:
                psi *= 2 * numpy.pi * L / K
            return psi

        def kernelvert(x, a, b, H):
            """
            integration kernel with vertical integration

            Parameters
            ----------
            x :     float or array-like
                independent parameter of the function
            a :     float
                lateral correlation parameter
            b :     complex
                vertical correlation parameter
            H :     float
                Hurst factor (scalar)

            Returns
            -------
            float or arraylike
            """
            w = numpy.exp(b * numpy.exp(-x**(2*H))) - 1
            F = 2 * numpy.cos(a*x) * w
            return F

        def kernel(x, a, b, H):
            """
            integration kernel without vertical integration

            Parameters
            ----------
            x :     float or array-like
                independent parameter of the function
            a :     float
                lateral correlation parameter
            b :     complex
                vertical correlation parameter
            H :     float
                Hurst factor (scalar)

            Returns
            -------
            float or arraylike
            """
            w = numpy.exp(b * numpy.exp(-x**(2*H))) - 1
            F = x * j0(a*x) * w
            return F

        def cquad(func, a, b, **kwargs):
            """
            complex quadrature by spliting real and imaginary part using scipy
            """
            def real_func(*args):
                return numpy.real(func(*args))

            def imag_func(*args):
                return numpy.imag(func(*args))
            real_integral = integrate.quad(real_func, a, b, **kwargs)
            imag_integral = integrate.quad(imag_func, a, b, **kwargs)
            return (real_integral[0] + 1j*imag_integral[0])

        # begin of _xrrdiffv2
        K = 2 * numpy.pi / lam

        N = len(thick)
        NqL = len(qL)
        Nqz = len(qz)

        QZ, QL = numpy.meshgrid(qz, qL)

        # scan types:
        if scan == 1:  # coplanar geometry
            Q = numpy.sqrt(QL**2 + QZ**2)
            QP = numpy.abs(QL)
            th = numpy.arcsin(Q / 2 / K)
            om = numpy.arctan2(QL, QZ)
            ALPHAI = th + om
            ALPHAF = th - om
        elif scan == 2:  # GISAXS geometry with constant incidence angle
            ALPHAI = numpy.radians(alphai) * numpy.ones((NqL, Nqz))
            ALPHAF = numpy.arcsin(QZ / K - numpy.sin(numpy.radians(alphai)))
            PHI = numpy.arcsin(QL / K / numpy.cos(ALPHAF))
            QP = K * numpy.sqrt(numpy.cos(ALPHAF)**2 + numpy.cos(ALPHAI)**2 -
                                2*numpy.cos(ALPHAF) * numpy.cos(ALPHAI) *
                                numpy.cos(PHI))
        elif scan == 3:  # with quasi omega/2theta scan in GISAXS geometry
            ALPHAI = numpy.arcsin(QZ * (K - numpy.sqrt(K**2 - QL**2)) / QL**2)
            ALPHAF = numpy.arcsin(QZ / K - numpy.sin(ALPHAI))
            PHI = numpy.arcsin(QL / K / numpy.cos(ALPHAF))
            QP = K * numpy.sqrt(numpy.cos(ALPHAF)**2 + numpy.cos(ALPHAI)**2 -
                                2*numpy.cos(ALPHAF) * numpy.cos(ALPHAI) *
                                numpy.cos(PHI))
        else:
            raise ValueError("Invalid value of parameter 'scan'")

        # removing the values under the horizon
        isurf = heaviside(ALPHAI) * heaviside(ALPHAF)

        # non-disturbed states:
        if method == 1:
            k01 = -K * numpy.sin(ALPHAI)
            kz1 = -K * numpy.sqrt(numpy.sin(ALPHAI)**2 - 2*deltaA)
            k02 = -K * numpy.sin(ALPHAF)
            kz2 = -K * numpy.sqrt(numpy.sin(ALPHAF)**2 - 2*deltaA)
            T1 = 2 * k01 / (k01 + kz1)
            T2 = 2 * k02 / (k02 + kz2)
            R01 = (k01 - kz1) / (k01 + kz1)
            R02 = (k02 - kz2) / (k02+kz2)
            R1 = numpy.zeros((NqL, Nqz), dtype=numpy.complex)
            R2 = numpy.copy(R1)
            nproc = 1
        else:  # method == 2
            T1, R1, R01, k01, kz1 = coherent(ALPHAI, K, delta, thick, N,
                                             NqL, Nqz)
            T2, R2, R02, k02, kz2 = coherent(ALPHAF, K, delta, thick, N,
                                             NqL, Nqz)
            nproc = 4

        # sample surface
        if beamwidth > 0:
            S = samplewidth * numpy.sin(ALPHAI) / beamwidth
            S[S > 1] = 1
        else:
            S = 1

        # z-coordinates
        z = numpy.zeros(N+1)
        for jn in range(1, N+1):
            z[jn] = z[jn-1] - thick[jn-1]

        # calculation of the deltas
        delt = numpy.zeros(N+1, dtype=numpy.complex)
        for jn in range(N+1):
            if jn == 0:
                delt[jn] = delta[jn]
            if jn > 0:
                delt[jn] = delta[jn] - delta[jn-1]

        # double sum over interfaces
        result = numpy.zeros((NqL, Nqz))
        for jn in range(N+1):
            # if method == 1 and (H == 1 or H == 0.5):
            #     print(jn)
            if nu != 0 or xiV == 0:
                jmdol = 1
            else:
                jmdol = numpy.argmin(numpy.abs(z - (z[jn] -
                                               xiV * numpy.log(eps))))

            for ja in range(nproc):
                if method == 1:
                    Qn = -kz1 - kz2
                    An = T1 * T2 * numpy.exp(-1j*Qn*z[jn])
                else:  # method == 2
                    if ja == 0:
                        An = T1[jn, ...] * T2[jn, ...]
                        Qn = -kz1[jn, ...] - kz2[jn, ...]
                    elif ja == 1:
                        An = T1[jn, ...] * R2[jn, ...]
                        Qn = -kz1[jn, ...] + kz2[jn, ...]
                    elif ja == 2:
                        An = R1[jn, ...] * T2[jn, ...]
                        Qn = kz1[jn, ...] - kz2[jn, ...]
                    elif ja == 3:
                        An = R1[jn, ...] * R2[jn, ...]
                        Qn = kz1[jn, ...] + kz2[jn, ...]
                for jm in range(jmdol, jn+1):
                    if jm == jn:
                        weight = 1
                    else:
                        weight = 2
                    # if method == 1 and (H != 0.5 and H != 1) and ja==1:
                    #     print(jn, jm)
                    # vertical correlation function:
                    if xiV > 0:
                        CV = numpy.exp(-abs(z[jn] - z[jm]) *
                                       (QP/numpy.max(QP))**nu / xiV)
                    else:
                        CV = 1
                    # effective values of sigma and lateral correl. length:
                    try:
                        LP = ((float(xiL[jn])**(-2*H) +
                               float(xiL[jm])**(-2*H)) / 2) ** (-1 / 2 / H)
                    except ZeroDivisionError:
                        LP = 0
                    sig = pymath.sqrt(sigma[jn] * sigma[jm])
                    for jb in range(nproc):
                        if method == 1:
                            Qm = -kz1 - kz2
                            Am = T1 * T2 * numpy.exp(-1j * Qm * z[jm])
                        else:  # method == 2
                            # if H != 0.5 or H != 1:
                            #    print(ja, jb, jn, jm)
                            if jb == 0:
                                Am = T1[jm, ...] * T2[jm, ...]
                                Qm = -kz1[jm, ...] - kz2[jm, ...]
                            elif jb == 1:
                                Am = T1[jm, ...] * R2[jm, ...]
                                Qm = -kz1[jm, ...] + kz2[jm, ...]
                            elif jb == 2:
                                Am = R1[jm, ...] * T2[jm, ...]
                                Qm = kz1[jm, ...] - kz2[jm, ...]
                            elif jb == 3:
                                Am = R1[jm, ...] * R2[jm, ...]
                                Qm = +kz1[jm, ...] + kz2[jm, ...]
                        # lateral correlation function:
                        Psi = correl(QP*LP, Qn*numpy.conj(Qm)*sig**2, LP, H,
                                     eps, nmax, vert, K, NqL, Nqz, isurf)
                        result += numpy.real(CV * delt[jn] *
                                             numpy.exp(-Qn**2 *
                                                       sigma[jn]**2/2) /
                                             Qn * An *
                                             numpy.conj(delt[jm] *
                                             numpy.exp(-Qm**2*sigma[jm]**2/2) /
                                             Qm * Am) * Psi) * weight

        result[isurf == 0] = 0
        self._smap_R01 = R01 * isurf
        self._smap_R02 = R02 * isurf
        self._smap_alphai = numpy.degrees(ALPHAI*isurf)
        self._smap_alphaf = numpy.degrees(ALPHAF*isurf)
        return result * S * K**4 / (16*numpy.pi**2)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
module for reading ILL data files (station D23): numor files
"""

import collections.abc
import os.path
import re

import numpy

from ..exception import InputError
# relative imports from xrayutilities
from .helper import xu_open

re_comment = re.compile(r"^A+$")
re_basicinfo = re.compile(r"^R+$")
re_values = re.compile(r"^F+$")
re_spectrum = re.compile(r"^S+$")
re_header = re.compile(r"^I+$")


class numorFile(object):
    """
    Represents a ILL data file (numor). The file is read during the Constructor
    call. This class should work for created at station D23 using the mad
    acquisition system.

    Parameters
    ----------
    filename :  str
        a string with the name of the data file
    """

    columns = {0: ('detector', 'monitor', 'time', 'gamma', 'omega', 'psi'),
               1: ('detector', 'monitor', 'time', 'gamma'),
               2: ('detector', 'monitor', 'time', 'omega'),
               5: ('detector', 'monitor', 'time', 'psi')}

    def __init__(self, filename, path=None):
        """
        constructor for the data file parser

        Parameters
        ----------
        filename :  str
            a string with the name of the data file
        path :      str, optional
            directory of the data file
        """
        self.filename = filename
        if path is None:
            self.full_filename = self.filename
        else:
            self.full_filename = os.path.join(path, self.filename)

        self.Read()

    def getline(self, fid):
        return fid.readline().decode('ascii')

    def ssplit(self, string):
        """
        multispace split. splits string at two or more spaces after stripping
        it.
        """
        return re.split(r'\s\s+', string.strip())

    def Read(self):
        """
        Read the data from the file
        """
        with xu_open(self.full_filename) as fid:
            self.filesize = os.stat(self.full_filename).st_size
            # read header
            self.init_mopo = {}
            self.comments = []
            self.header = {}
            self._data = []
            while fid.tell() < self.filesize:
                line = self.getline(fid)

                if re_comment.match(line):
                    # read AAAA sections
                    line = self.getline(fid)
                    desc = []
                    for j in range(int(line.split()[1])):
                        desc += self.ssplit(self.getline(fid))
                    comval = self.ssplit(self.getline(fid))
                    self.comments.append((desc, comval))

                if re_basicinfo.match(line):
                    # read RRRR section
                    info = self.ssplit(self.getline(fid))
                    self.dataversion = int(info[2])
                    self.runnumber = int(info[0])

                    if int(info[1]) > 0:
                        headerdesc = ''
                        for j in range(int(info[1])):
                            headerdesc += self.getline(fid) + '\n'
                        self.comments.append((['Fileheader'], [headerdesc]))

                if re_header.match(line):
                    # read IIII section: integer header values
                    info = self.ssplit(self.getline(fid))
                    names = []
                    values = []

                    for j in range(int(info[1])):
                        names += self.getline(fid).split()
                    values = numpy.fromfile(fid, dtype=int,
                                            count=int(info[0]), sep=' ')
                    self.header = {k: v for k, v in zip(names, values)}

                if re_values.match(line):
                    # read FFFF section: initial motor positions
                    info = self.ssplit(self.getline(fid))
                    names = []
                    values = []

                    for j in range(int(info[1])):
                        names += self.ssplit(self.getline(fid))
                    values = numpy.fromfile(fid, dtype=float,
                                            count=int(info[0]), sep=' ')
                    self.init_mopo = {k: v for k, v in zip(names, values)}

                if re_spectrum.match(line):
                    # read SSSS section: initial motor positions
                    info = self.ssplit(self.getline(fid))
                    self.nspectra = int(info[2])

                    if re_values.match(self.getline(fid)):
                        # read FFFF section: subspectrum data
                        nval = int(self.getline(fid))
                        # check if nval is multiple of npdone
                        if nval % self.header['npdone'] != 0:
                            raise InputError("File corrupted, wrong number of "
                                             "data values (%d) found." % nval)

                        self._data.append(numpy.fromfile(fid, dtype=float,
                                                         count=nval, sep=' '))

                    if int(info[1]) == 0:
                        break
            # make data columns accessible by names
            data = numpy.reshape(self._data[0],
                                 (self.header['npdone'],
                                  nval // self.header['npdone']))
            self.data = numpy.rec.fromrecords(
                data, names=self.columns[self.header['manip']])

    def __str__(self):
        ostr = 'Numor: %d (%s)\n' % (self.runnumber, self.filename)
        ostr += 'Comments: %s\n' % " ".join(
            s for c in self.comments for s in c[1])
        ostr += 'Npoints/Ndone: %(nkmes)d/%(npdone)d\n' % (self.header)
        ostr += 'Nspectra: %d\n' % self.nspectra
        ostr += 'Ncolumns: %s' % self.data.shape[1]
        return ostr


def numor_scan(scannumbers, *args, **kwargs):
    """
    function to obtain the angular cooridinates as well as intensity values
    saved in numor datafiles. Especially useful for combining several scans
    into one data object.

    Parameters
    ----------
    scannumbers :   int or str or iterable
        number of the numors, or list of numbers. This will be transformed to a
        string and used as a filename
    args :          str, optional
        names of the motors e.g.: 'omega', 'gamma'
    kwargs :        dict
        keyword arguments are passed on to numorFile. e.g. 'path' for the files
        directory

    Returns
    -------
    [ang1, ang2, ...] :     list
        angular positions list, omitted if no args are given
    data :                  ndarray
        all the data values.

    Examples
    --------
    >>> [om, gam], data = xu.io.numor_scan(414363, 'omega', 'gamma')
    """

    if isinstance(scannumbers, (str, int)):
        scanlist = list([scannumbers])
    elif isinstance(scannumbers, collections.abc.Iterable):
        scanlist = scannumbers
    else:
        raise TypeError('scannumbers is of invalid type (%s)'
                        % type(scannumbers))

    angles = dict.fromkeys(args)
    for key in angles:
        if not isinstance(key, str):
            raise InputError("*arg values need to be strings with motornames")
        angles[key] = numpy.zeros(0)
    buf = numpy.zeros(0)
    MAP = numpy.zeros(0)

    for nr in scanlist:
        scan = numorFile(str(nr), **kwargs)
        sdata = scan.data
        if MAP.dtype == numpy.float64:
            MAP.dtype = sdata.dtype
        # append scan data to MAP, where all data are stored
        MAP = numpy.append(MAP, sdata)
        # check type of scan
        for i in range(len(args)):
            motname = args[i]
            scanlength = len(sdata)
            try:
                buf = sdata[motname]
            except ValueError:
                mv = [v for k, v in scan.init_mopo.items()
                      if motname in k][0]
                buf = mv * numpy.ones(scanlength)
            angles[motname] = numpy.concatenate((angles[motname], buf))

    retval = []
    for motname in args:
        # create return values in correct order
        retval.append(angles[motname])

    if not args:
        return MAP
    elif len(args) == 1:
        return retval[0], MAP
    else:
        return retval, MAP



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2013-2017 Dominik Kriegner <dominik.kriegner@gmail.com>

import numpy

from . import elements as e
from .heuslerlib import *
from .material import (Crystal, CubicAlloy, CubicElasticTensor,
                       HexagonalElasticTensor, WZTensorFromCub)
from .spacegrouplattice import SGLattice

# some predefined materials
# PLEASE use N/m^2 as unit for cij for newly entered material ( 1 dyn/cm^2 =
# 0.1 N/m^2 = 0.1 GPa)
# Use Kelvin as unit for the Debye temperature
# ref http://www.semiconductors.co.uk/propiviv5431.htm
C = Crystal("C", SGLattice('227:1', 3.5668, atoms=[e.C, ], pos=['8a', ]),
            CubicElasticTensor(1.0764e12, 0.125e12, 0.577e12))
C_HOPG = Crystal('HOPG', SGLattice(194, 2.4612, 6.7079,
                                   atoms=[e.C, e.C], pos=['2b', '2c']))
Si = Crystal("Si", SGLattice('227:1', 5.43104, atoms=[e.Si, ], pos=['8a', ]),
             CubicElasticTensor(165.77e+9, 63.93e+9, 79.62e+9), thetaDebye=640)
Ge = Crystal("Ge", SGLattice('227:1', 5.65785, atoms=[e.Ge, ], pos=['8a', ]),
             CubicElasticTensor(128.5e+9, 48.3e+9, 66.8e+9), thetaDebye=374)
InAs = Crystal("InAs", SGLattice(216, 6.0583, atoms=[e.In, e.As],
                                 pos=['4a', '4c']),
               CubicElasticTensor(8.34e+10, 4.54e+10, 3.95e+10),
               thetaDebye=280)
InP = Crystal("InP", SGLattice(216, 5.8687, atoms=[e.In, e.P],
                               pos=['4a', '4c']),
              CubicElasticTensor(10.11e+10, 5.61e+10, 4.56e+10),
              thetaDebye=425)
InSb = Crystal("InSb", SGLattice(216, 6.47937, atoms=[e.In, e.Sb],
                                 pos=['4a', '4c']),
               CubicElasticTensor(6.66e+10, 3.65e+10, 3.02e+10),
               thetaDebye=160)
GaP = Crystal("GaP", SGLattice(216, 5.4505, atoms=[e.Ga, e.P],
                               pos=['4a', '4c']),
              CubicElasticTensor(14.05e+10, 6.20e+10, 7.03e+10),
              thetaDebye=445)
GaAs = Crystal("GaAs", SGLattice(216, 5.65325, atoms=[e.Ga, e.As],
                                 pos=['4a', '4c']),
               CubicElasticTensor(11.9e+10, 5.34e+10, 5.96e+10),
               thetaDebye=360)
AlAs = Crystal("AlAs", SGLattice(216, 5.6611, atoms=[e.Al, e.As],
                                 pos=['4a', '4c']),
               CubicElasticTensor(12.02e+10, 5.70e+10, 5.99e+10),
               thetaDebye=446)
GaSb = Crystal("GaSb", SGLattice(216, 6.09593, atoms=[e.Ga, e.Sb],
                                 pos=['4a', '4c']),
               CubicElasticTensor(8.83e+10, 4.02e+10, 4.32e+10),
               thetaDebye=266)
# from Cryst. Growth Des. 15, 4795-4803 (2015)
GaAsWZ = Crystal("GaAs(WZ)",
                 SGLattice(186, 3.9845, 6.5701, atoms=[e.Ga, e.As],
                           pos=[('2b', 0), ('2b', 3/8.)]),
                 WZTensorFromCub(11.9e+10, 5.34e+10, 5.96e+10))
GaAs4H = Crystal("GaAs(4H)",
                 SGLattice(186, 3.9900, 13.0964,
                           atoms=[e.Ga, e.Ga, e.As, e.As],
                           pos=[('2a', 0), ('2b', 1/4.),
                                ('2a', 3/16.), ('2b', 7/16.)]))
# from Phys. Rev. B 88, 115315 (2013)
GaPWZ = Crystal("GaP(WZ)",
                SGLattice(186, 3.8419, 6.3353, atoms=[e.Ga, e.P],
                          pos=[('2b', 0), ('2b', 0.37385)]),
                WZTensorFromCub(14.05e+10, 6.20e+10, 7.03e+10))
# from Nanotechnology 22 425704 (2011)
InPWZ = Crystal("InP(WZ)",
                SGLattice(186, 4.1423, 6.8013, atoms=[e.In, e.P],
                          pos=[('2b', 0), ('2b', 3/8.)]),
                WZTensorFromCub(10.11e+10, 5.61e+10, 4.56e+10))
# from Nano Lett., 2011, 11 (4), pp 1483-1489
InAsWZ = Crystal("InAs(WZ)",
                 SGLattice(186, 4.2742, 7.0250, atoms=[e.In, e.As],
                           pos=[('2b', 0), ('2b', 3/8.)]),
                 WZTensorFromCub(8.34e+10, 4.54e+10, 3.95e+10))
InAs4H = Crystal("InAs(4H)",
                 SGLattice(186, 4.2780, 14.0171,
                           atoms=[e.In, e.In, e.As, e.As],
                           pos=[('2a', 0), ('2b', 1/4.),
                                ('2a', 3/16.), ('2b', 7/16.)]))
InSbWZ = Crystal("InSb(WZ)",
                 SGLattice(186, 4.5712, 7.5221, atoms=[e.In, e.Sb],
                           pos=[('2b', 0), ('2b', 3/8.)]),
                 WZTensorFromCub(6.66e+10, 3.65e+10, 3.02e+10))
InSb4H = Crystal("InSb(4H)",
                 SGLattice(186, 4.5753, 15.0057,
                           atoms=[e.In, e.In, e.Sb, e.Sb],
                           pos=[('2a', 0), ('2b', 1/4.),
                                ('2a', 3/16.), ('2b', 7/16.)]))

# ? Unit of elastic constants for CdTe, PbTe, PbSe ?
PbTe = Crystal("PbTe",
               SGLattice(225, 6.464, atoms=[e.Pb, e.Te], pos=['4a', '4b']),
               CubicElasticTensor(93.6, 7.7, 13.4))
PbSe = Crystal("PbSe",
               SGLattice(225, 6.128, atoms=[e.Pb, e.Se], pos=['4a', '4b']),
               CubicElasticTensor(123.7, 19.3, 15.9))
CdTe = Crystal("CdTe", SGLattice(216, 6.482, atoms=[e.Cd, e.Te],
                                 pos=['4a', '4c']),
               CubicElasticTensor(53.5, 36.7, 19.9))
CdSe = Crystal("CdSe", SGLattice(186, 4.5712, 7.5221, atoms=[e.In, e.Sb],
                                 pos=[('2b', 0), ('2b', 3/8.)]),
               HexagonalElasticTensor(7.490e10, 4.609e10, 3.926e10,
                                      8.451e10, 1.315e10))
CdSe_ZB = Crystal("CdSe(ZB)", SGLattice(216, 6.052, atoms=[e.Cd, e.Se],
                                        pos=['4a', '4c']))
HgSe = Crystal("HgSe", SGLattice(216, 6.085, atoms=[e.Hg, e.Se],
                                 pos=['4a', '4c']),
               CubicElasticTensor(6.1e10, 4.4e10, 2.2e10))
NaCl = Crystal("NaCl",
               SGLattice(225, 5.6402, atoms=[e.Na, e.Cl], pos=['4a', '4b']))
MgO = Crystal("MgO",
              SGLattice(225, 4.212, atoms=[e.Mg, e.O], pos=['4a', '4b']))
GaN = Crystal("GaN",
              SGLattice(186, 3.189, 5.186, atoms=[e.Ga, e.N],
                        pos=[('2b', 0), ('2b', 3/8.)]),
              HexagonalElasticTensor(390.e9, 145.e9, 106.e9, 398.e9, 105.e9),
              thetaDebye=600)
BaF2 = Crystal("BaF2", SGLattice(225, 6.2001, atoms=[e.Ba, e.F],
                                 pos=['4a', '8c']))
SrF2 = Crystal("SrF2", SGLattice(225, 5.8007, atoms=[e.Sr, e.F],
                                 pos=['4a', '8c']))
CaF2 = Crystal("CaF2", SGLattice(225, 5.4631, atoms=[e.Ca, e.F],
                                 pos=['4a', '8c']))
MnO = Crystal("MnO", SGLattice(225, 4.444, atoms=[e.Mn, e.O],
                               pos=['4a', '4b']))
MnTe = Crystal("MnTe", SGLattice(186, 4.1429, 6.7031, atoms=[e.Mn, e.Te],
                                 pos=[('2a', 0), ('2b', 0.25)]))
GeTe = Crystal("GeTe",
               SGLattice('160:R', 5.996, 88.18, atoms=[e.Ge, e.Ge, e.Te, e.Te],
                         pos=[('1a', -0.237), ('3b', (0.5-0.237, -0.237)),
                              ('1a', 0.237), ('3b', (0.5+0.237, +0.237))]))
SnTe = Crystal("SnTe",
               SGLattice(225, 6.3268, atoms=[e.Sn, e.Te], pos=['4a', '4b']))
Al = Crystal("Al", SGLattice(225, 4.04958, atoms=[e.Al, ], pos=['4a', ]))
Au = Crystal("Au", SGLattice(225, 4.0782, atoms=[e.Au, ], pos=['4a', ]))
Fe = Crystal("Fe", SGLattice(229, 2.8665, atoms=[e.Fe, ], pos=['2a', ]))
Cr = Crystal("Cr", SGLattice(229, 2.910, atoms=[e.Cr, ], pos=['2a', ]))
Co = Crystal("Co", SGLattice(194, 2.5071, 4.0695, atoms=[e.Co, ],
                             pos=['2c', ]))
Ti = Crystal("Ti", SGLattice(194, 2.9508, 4.6855, atoms=[e.Ti, ],
                             pos=['2c', ]))
Mo = Crystal("Mo", SGLattice(229, 3.147, atoms=[e.Mo, ], pos=['2a', ]))
Ru = Crystal("Ru", SGLattice(194, 2.7059, 4.2815, atoms=[e.Ru, ],
                             pos=['2c', ]))
Rh = Crystal("Rh", SGLattice(225, 3.8034, atoms=[e.Rh, ], pos=['4a', ]))
V = Crystal("V", SGLattice(229, 3.024, atoms=[e.V, ], pos=['2a', ]))
Ta = Crystal("Ta", SGLattice(229, 3.306, atoms=[e.Ta, ], pos=['2a', ]))
Nb = Crystal("Nb", SGLattice(229, 3.3004, atoms=[e.Nb, ], pos=['2a', ]))
Pt = Crystal("Pt", SGLattice(225, 3.9242, atoms=[e.Pt, ], pos=['4a', ]))
Ag2Se = Crystal("Ag2Se", SGLattice(19, 4.333, 7.062, 7.764,
                                   atoms=[e.Ag, e.Ag, e.Se],
                                   pos=[('4a', (0.107, 0.369, 0.456)),
                                        ('4a', (0.728, 0.029, 0.361)),
                                        ('4a', (0.358, 0.235, 0.149))]))
TiO2 = Crystal("TiO2", SGLattice(136, 4.59, 2.96, atoms=[e.Ti, e.O],
                                 pos=['2a', ('4f', 0.30479)]))
MnO2 = Crystal("MnO2", SGLattice(136, 4.40, 2.87, atoms=[e.Mn, e.O],
                                 pos=['2a', ('4f', 0.30479)]))
VO2_Rutile = Crystal("VO2", SGLattice(136, 4.55, 2.88, atoms=[e.V, e.O],
                                      pos=['2a', ('4f', 0.305)]))
VO2_Baddeleyite = Crystal("VO2", SGLattice(14, 5.75, 5.42, 5.38, 122.6,
                                           atoms=[e.V, e.O, e.O],
                                           pos=[('4e', (0.242, 0.975, 0.025)),
                                                ('4e', (0.1, 0.21, 0.20)),
                                                ('4e', (0.39, 0.69, 0.29))]))
SiO2 = Crystal("SiO2", SGLattice(154, 4.916, 5.4054, atoms=[e.Si, e.O],
                                 pos=[('3a', 0.46970),
                                      ('6c', (0.41350, 0.26690,
                                              0.11910+2/3.))]))
In = Crystal("In", SGLattice(139, 3.2523, 4.9461, atoms=[e.In, ],
                             pos=['2a', ]))
Sb = Crystal("Sb", SGLattice('166:H', 4.307, 11.273, atoms=[e.Sb, ],
                             pos=[('6c', 0.23349), ]))
Sn = Crystal("Sn", SGLattice('141:1', 5.8197, 3.17488, atoms=[e.Sn, ],
                             pos=['4a', ]))
Ag = Crystal("Ag", SGLattice(225, 4.0853, atoms=[e.Ag, ], pos=['4a', ]))
SnAlpha = Crystal("Sn-alpha", SGLattice('227:1', 6.4912, atoms=[e.Sn, ],
                                        pos=['8a', ]))
Cu = Crystal("Cu", SGLattice(225, 3.61496, atoms=[e.Cu, ], pos=['4a', ]))
CaTiO3 = Crystal("CaTiO3", SGLattice(221, 3.795, atoms=[e.Ca, e.Ti, e.O],
                                     pos=['1a', '1b', '3c']))
SrTiO3 = Crystal("SrTiO3", SGLattice(221, 3.905, atoms=[e.Sr, e.Ti, e.O],
                                     pos=['1a', '1b', '3c']))
BaTiO3 = Crystal("BaTiO3", SGLattice(99, 3.992, 4.036,
                                     atoms=[e.Ba, e.Ti, e.O, e.O],
                                     pos=[('1a', 1.000), ('1b', 0.5274),
                                          ('1b', 0.9993), ('2c', 0.5125)]))
# BiFeO3 = Crystal("BiFeO3", SGLattice())
# BiFeO3 = Crystal(
#    "BiFeO3",
#    lattice.PerovskiteTypeRhombohedral(elements.Bi, elements.Fe, elements.O,
#                                       3.965, 89.3))
FeO = Crystal("FeO", SGLattice(225, 4.332, atoms=[e.Fe, e.O],
                               pos=['4a', '4b']))
CoO = Crystal("CoO", SGLattice(225, 4.214, atoms=[e.Co, e.O],
                               pos=['4a', '4b']))
Fe3O4 = Crystal("Fe3O4", SGLattice('227:2', 8.3958, atoms=[e.Fe, e.Fe, e.O],
                                   pos=['8a', '16d', ('32e', 0.255)]))
Co3O4 = Crystal("Co3O4", SGLattice('227:2', 8.0821, atoms=[e.Co, e.Co, e.O],
                                   pos=['8a', '16d', ('32e', 0.255)]))
FeRh = Crystal("FeRh", SGLattice(221, 2.993, atoms=[e.Fe, e.Rh],
                                 pos=['1a', '1b']))
Ir20Mn80 = Crystal("Ir20Mn80", SGLattice(225, 3.780, atoms=[e.Ir, e.Mn],
                                         pos=['4a', '4a'], occ=[0.2, 0.8]))
CoFe = Crystal("CoFe", SGLattice(221, 2.8508, atoms=[e.Co, e.Fe],
                                 pos=['1a', '1b']))
CoGa = Crystal("CoGa", SGLattice(221, 2.883, atoms=[e.Co, e.Ga],
                                 pos=['1a', '1b']))
LaB6 = Crystal("LaB6", SGLattice(221, 4.15692, atoms=[e.La, e.B],
                                 pos=['1a', ('6f', 0.19750)]))
Al2O3 = Crystal("Al2O3", SGLattice('167:H', 4.7602, 12.9933,
                                   atoms=[e.Al, e.O],
                                   pos=[('12c', 0.35216), ('18e', 0.30624)]))
CuMnAs = Crystal("CuMnAs", SGLattice('129:2', 3.8200, 6.3180,
                                     atoms=[e.Cu, e.Mn, e.As],
                                     pos=['2b', ('2c', -0.8300),
                                          ('2c', -0.2347)],
                                     occ=[1, 0.86, 0.96],
                                     b=[3.5531, 1.8950, 1.2633]))
Mn3Ge_cub = Crystal("Mn3Ge (cub)", SGLattice('221', 3.8019,
                                             atoms=[e.Mn, e.Ge],
                                             pos=['3c', '1a']))
Mn3Ge = Crystal("Mn3Ge (hex)", SGLattice('194', 5.34, 4.31,
                                         atoms=[e.Mn, e.Ge],
                                         pos=[('6h', 1/6.), '2d']))
Pt3Cr = Crystal("Pt3Cr", SGLattice('221', 3.876,
                                   atoms=[e.Pt, e.Cr], pos=['3c', '1a']))
TiN = Crystal("TiN",
              SGLattice(225, 4.235, atoms=[e.Ti, e.N], pos=['4a', '4b']))


# Alloys with special properties
class SiGe(CubicAlloy):

    def __init__(self, x):
        """
        Si_{1-x} Ge_x cubic compound
        """
        super().__init__(Si, Ge, x)

    @staticmethod
    def lattice_const_AB(latA, latB, x, **kwargs):
        """
        method to calculate the lattice parameter of the SiGe alloy with
        composition Si_{1-x}Ge_x
        """
        return latA + (0.2 * x + 0.027 * x ** 2) * \
            (latB - latA) / numpy.linalg.norm(latB - latA)


class AlGaAs(CubicAlloy):

    def __init__(self, x):
        """
        Al_{1-x} Ga_x As cubic compound
        """
        super().__init__(AlAs, GaAs, x)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2010 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
parser for the alignment log file of the rotating anode
"""

import re

import numpy

from .. import config, utilities
from .helper import xu_open

LOG_comment = re.compile(r"^#C")
LOG_peakname = re.compile(r"^#P")
LOG_motorname = re.compile(r"^#M")
LOG_datetime = re.compile(r"^#D")
LOG_tagline = re.compile(r"^#")
# denotes a numeric value
LOG_num_value = re.compile(r"[+-]*\d*\.*\d*e*[+-]*\d+")


class RA_Alignment(object):

    """
    class to parse the data file created by the alignment routine
    (tpalign) at the rotating anode spec installation

    this routine does an iterative alignment procedure and saves the
    center of mass values were it moves after each scan. It iterates
    between two different peaks and iteratively aligns at each peak between
    two different motors (om/chi at symmetric peaks, om/phi at asymmetric
    peaks)
    """

    def __init__(self, filename):
        """
        initialization function to initialize the objects variables and
        opens the file

        Parameters
        ----------
        filename :  str
            filename of the alignment log file
        """

        self.filename = filename
        try:
            self.fid = xu_open(self.filename)
        except OSError:
            self.fid = None
            raise IOError("error opening alignment log file %s"
                          % self.filename)

        self.peaks = []
        self.alignnames = []
        self.motorpos = []
        self.intensities = []
        self.iterations = []

        self.Parse()

    def Parse(self):
        """
        parser to read the alignment log and obtain the aligned values
        at every iteration.
        """

        currentpeakname = None
        currentmotname = None
        opencommenttag = False
        dataline = False
        iteration = 0

        if self.fid is None:
            raise Exception("RA_Alignment: file was not opened by "
                            "initialization!")

        for line in self.fid.readlines():
            # for loop to read every line in the file
            line = line.decode('ascii')

            # check for new tag in the current line
            if LOG_tagline.match(line):
                opencommenttag = False

                if LOG_comment.match(line):
                    # comment line or block starts
                    opencommenttag = True
                    continue

                elif LOG_datetime.match(line):
                    # data is so far ignored
                    continue

                elif LOG_peakname.match(line):
                    # line with peak name found
                    pname = LOG_peakname.sub("", line)
                    pname = pname.strip()
                    # check if we found a new peakname
                    try:
                        self.peaks.index(pname)
                    except ValueError:
                        self.peaks.append(pname)
                    currentpeakname = pname  # set current peak name
                    iteration += 1  # increment iteration counter

                elif LOG_motorname.match(line):
                    # line with motorname is found
                    motname = LOG_motorname.sub("", line)
                    motname = motname.strip()
                    # check if a peakname is already set
                    if currentpeakname is None:
                        if config.VERBOSITY >= config.INFO_LOW:
                            print("RA_Alignment: Warning: a peakname should "
                                  "be given before a motor data line")
                        currentpeakname = "somepeak"
                    currentmotname = currentpeakname + "_" + motname
                    # check if we found a new peak/motor name combination
                    try:
                        self.alignnames.index(currentmotname)
                    except ValueError:
                        # new peak/motor combination
                        self.alignnames.append(currentmotname)
                        # create necessary data structures
                        self.motorpos.append([])
                        self.intensities.append([])
                        self.iterations.append([])
                    # next line contains motor position and intensity
                    dataline = True
            elif opencommenttag:
                # ignore line because it is part of a comment block
                continue

            elif dataline:
                # dataline with motorposition and intensity is found
                line_list = LOG_num_value.findall(line)
                idx = self.alignnames.index(currentmotname)
                self.motorpos[idx].append(float(line_list[0]))
                self.intensities[idx].append(float(line_list[1]))
                self.iterations[idx].append(iteration)
                dataline = False

        # convert data to numpy array and combine position and intensity
        self.data = []
        for i, k in enumerate(self.keys()):
            self.data.append(numpy.array((self.motorpos[i],
                                          self.intensities[i],
                                          self.iterations[i])))

    def __str__(self):
        """
        returns a string describing the content of the alignment file
        """
        ostr = ""
        ostr += "Peaknames: " + repr(self.peaks) + "\n"
        ostr += "aligned values: " + repr(self.alignnames)
        return ostr

    def __del__(self):
        try:
            self.fid.close()
        except AttributeError:
            pass

    def keys(self):
        """
        returns a list of keys for which aligned values were parsed
        """
        return self.alignnames

    def get(self, key):
        return self.__getitem__(key)

    def __getitem__(self, key):
        """
        returns the values to the corresponding key
        """
        if key in self.alignnames:
            i = self.alignnames.index(key)
            return self.data[i]
        else:
            raise KeyError("RA_Alignment: unknown key given!")

    def plot(self, pname):
        """
        function to plot the alignment history for a given peak

        Parameters
        ----------
        pname :     str
            peakname for which the alignment should be plotted
        """
        flag, plt = utilities.import_matplotlib_pyplot('XU.io.RA_ALignment')
        if not flag:
            return

        if pname not in self.peaks:
            print("RA_Alignment.plot: error peakname not found!")
            return

        # get number aligned axis for the current peak
        axnames = []
        for k in self.keys():
            if k.find(pname) >= 0:
                axnames.append(k)

        fig, ax = plt.subplots(nrows=len(axnames), sharex=True)

        for an, axis in zip(axnames, ax):
            d = self.get(an)
            plt.sca(axis)
            plt.plot(d[2], d[0], 'k.-')
            plt.ylabel(re.sub(pname + "_", "", an))
            axis.twinx()
            plt.plot(d[2], d[1], 'r.-')
            plt.ylabel("Int (cps)", color='r')
            plt.grid()

        plt.xlabel("Peak iteration number")
        plt.suptitle(pname)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2019 Dominik Kriegner <dominik.kriegner@gmail.com>

import math

import numpy


def mosaic_analytic(qx, qz, RL, RV, Delta, hx, hz, shape):
    """
    simulation of the coplanar reciprocal space map of a single mosaic layer
    using a simple analytic approximation

    Parameters
    ----------
    qx :    array-like
        vector of the qx values (offset from the Bragg peak)
    qz :    array-like
        vector of the qz values (offset from the Bragg peak)
    RL :    float
        lateral block radius in Angstrom
    RV :    float
        vertical block radius in Angstrom
    Delta : float
        root mean square misorientation of the grains in degree
    hx :    float
        lateral component of the diffraction vector
    hz :    float
        vertical component of the diffraction vector
    shape:  float
        shape factor (1..Gaussian)

    Returns
    -------
    array-like
    2D array with calculated intensities
    """
    QX, QZ = numpy.meshgrid(qx, qz)
    QX = QX.T
    QZ = QZ.T
    DD = numpy.radians(Delta)
    tmp = 6 + DD**2 * ((hz*RL)**2 + (hx*RV)**2)
    F = ((DD*RL*RV)**2*(QZ*hz + QX*hx)**2+6*((RL*QX)**2+(RV*QZ)**2)) / 4 / tmp
    return math.pi * math.sqrt(6) * RL * RV /\
        math.sqrt(tmp) * numpy.exp(-F**shape)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012-2018 Dominik Kriegner <dominik.kriegner@gmail.com>
"""
script to create the HDF5 database from the raw data of XOP
his file is only needed for administration and also used during the unit tests

Optionally it accepts two command line arguments. The first one is the filename
of the database. If the filename has no path or a relative path the data-folder
given as second argument will be used as root of the output file.

The second optional argument is the directory of the source data file used to
generate the database. If no path is given the 'data' sub-folder of the scripts
directory is used. The script therefore works if called from the extracted
tarball directory but not after installation. After installation the respective
data path must be specified. The script can be run with 0, one or two command
line arguments as described above.

See tests/test_materials_database for an example.
"""

import lzma
import os.path
import sys

# local import
from database import (DataBase, add_color_from_JMOL, add_f0_from_intertab,
                      add_f1f2_from_kissel, add_mass_from_NIST,
                      add_radius_from_WIKI, init_material_db)

verbose = False

if len(sys.argv) > 1:
    fname = sys.argv[1]
if len(sys.argv) > 2:
    dataroot = sys.argv[2]
else:
    dataroot = os.path.join(os.path.dirname(__file__), 'data')

dbf = DataBase(fname)
dbf.Create('elementdata',
           'Database with elemental data from XOP and Kissel databases')

init_material_db(dbf)

# add a dummy element, this is useful not only for testing and should be
# kept in future! It can be used for structure factor calculation tests, and
# shows how the a database entry can be generated manually
dbf.SetMaterial('dummy')
dbf.SetF0([1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0])  # atomic structure factors
dbf.SetF1F2((0, 1e5), (0, 0), (0, 0))  # zero dispersion correction

add_mass_from_NIST(dbf, os.path.join(dataroot, 'nist_atom.dat'), verbose)
add_color_from_JMOL(dbf, os.path.join(dataroot, 'colors.dat'), verbose)
add_radius_from_WIKI(dbf, os.path.join(dataroot, 'atomic_radius.dat'), verbose)

# add F0(Q) for every element
# with lzma.open(os.path.join('data', 'f0_xop.dat.xz'), 'r') as xop:
#    add_f0_from_xop(dbf, xop, verbose)
with lzma.open(os.path.join(dataroot, 'f0_InterTables.dat.xz'), 'r') as itf:
    add_f0_from_intertab(dbf, itf, verbose)

# add F1 and F2 from database
with lzma.open(os.path.join(dataroot, 'f1f2_asf_Kissel.dat.xz'), 'r') as kf:
    add_f1f2_from_kissel(dbf, kf, verbose)
# with lzma.open(os.path.join(dataroot, 'f1f2_Henke.dat'), 'r') as hf:
#    add_f1f2_from_henkedb(dbf, hf, verbose)

# Also its possible to add custom data from different databases; e.g.
# created by Hepaestus (http://bruceravel.github.io/demeter/). This is also
# possible for specific elements only, therefore extract the data from
# Hephaestus or any other source producing ASCII files with three columns
# (energy (eV), f1, f2). To import such data use:
# add_f1f2_from_ascii_file(dbf, os.path.join(dataroot, 'Ga.f1f2'), 'Ga',
#                          verbose)

dbf.Close()



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2009 Eugen Wintersberger <eugen.wintersberger@desy.de>
# Copyright (C) 2010-2015 Dominik Kriegner <dominik.kriegner@gmail.com>

from .atom import Atom

dummy = Atom('dummy', 0)
H = Atom('H', 1)
Hdot = Atom('H.', 1)
H1m = Atom('H1-', 1)
D = Atom('D', 1)
T = Atom('T', 1)
He = Atom('He', 2)
Li = Atom('Li', 3)
Li1p = Atom('Li1+', 3)
Be = Atom('Be', 4)
Be2p = Atom('Be2+', 4)
B = Atom('B', 5)
C = Atom('C', 6)
Cdot = Atom('C.', 6)
N = Atom('N', 7)
O = Atom('O', 8)
O1m = Atom('O1-', 8)
O2mdot = Atom('O2-.', 8)
O2m = Atom('O2-.', 8)
F = Atom('F', 9)
F1m = Atom('F1-', 9)
Ne = Atom('Ne', 10)
Na = Atom('Na', 11)
Na1p = Atom('Na1+', 11)
Mg = Atom('Mg', 12)
Mg2p = Atom('Mg2+', 12)
Al = Atom('Al', 13)
Al3p = Atom('Al3+', 13)
Si = Atom('Si', 14)
Sidot = Atom('Si.', 14)
Si4p = Atom('Si4+', 14)
P = Atom('P', 15)
S = Atom('S', 16)
Cl = Atom('Cl', 17)
Cl1m = Atom('Cl1-', 17)
Ar = Atom('Ar', 18)
K = Atom('K', 19)
K1p = Atom('K1+', 19)
Ca = Atom('Ca', 20)
Ca2p = Atom('Ca2+', 20)
Sc = Atom('Sc', 21)
Sc3p = Atom('Sc3+', 21)
Ti = Atom('Ti', 22)
Ti2p = Atom('Ti2+', 22)
Ti3p = Atom('Ti3+', 22)
Ti4p = Atom('Ti4+', 22)
V = Atom('V', 23)
V2p = Atom('V2+', 23)
V3p = Atom('V3+', 23)
V5p = Atom('V5+', 23)
Cr = Atom('Cr', 24)
Cr2p = Atom('Cr2+', 24)
Cr3p = Atom('Cr3+', 24)
Mn = Atom('Mn', 25)
Mn2p = Atom('Mn2+', 25)
Mn3p = Atom('Mn3+', 25)
Mn4p = Atom('Mn4+', 25)
Fe = Atom('Fe', 26)
Fe2p = Atom('Fe2+', 26)
Fe3p = Atom('Fe3+', 26)
Co = Atom('Co', 27)
Co2p = Atom('Co2+', 27)
Co3p = Atom('Co3+', 27)
Ni = Atom('Ni', 28)
Ni2p = Atom('Ni2+', 28)
Ni3p = Atom('Ni3+', 28)
Cu = Atom('Cu', 29)
Cu1p = Atom('Cu1+', 29)
Cu2p = Atom('Cu2+', 29)
Zn = Atom('Zn', 30)
Zn2p = Atom('Zn2+', 30)
Ga = Atom('Ga', 31)
Ga3p = Atom('Ga3+', 31)
Ge = Atom('Ge', 32)
Ge4p = Atom('Ge4+', 32)
As = Atom('As', 33)
Se = Atom('Se', 34)
Br = Atom('Br', 35)
Br1m = Atom('Br1-', 35)
Kr = Atom('Kr', 36)
Rb = Atom('Rb', 37)
Rb1p = Atom('Rb1+', 37)
Sr = Atom('Sr', 38)
Sr2p = Atom('Sr2+', 38)
Y = Atom('Y', 39)
Y3p = Atom('Y3+', 39)
Zr = Atom('Zr', 40)
Zr4p = Atom('Zr4+', 40)
Nb = Atom('Nb', 41)
Nb3p = Atom('Nb3+', 41)
Nb5p = Atom('Nb5+', 41)
Mo = Atom('Mo', 42)
Mo3p = Atom('Mo3+', 42)
Mo5p = Atom('Mo5+', 42)
Mo6p = Atom('Mo6+', 42)
Tc = Atom('Tc', 43)
Ru = Atom('Ru', 44)
Ru3p = Atom('Ru3+', 44)
Ru4p = Atom('Ru4+', 44)
Rh = Atom('Rh', 45)
Rh3p = Atom('Rh3+', 45)
Rh4p = Atom('Rh4+', 45)
Pd = Atom('Pd', 46)
Pd2p = Atom('Pd2+', 46)
Pd4p = Atom('Pd4+', 46)
Ag = Atom('Ag', 47)
Ag1p = Atom('Ag1+', 47)
Ag2p = Atom('Ag2+', 47)
Cd = Atom('Cd', 48)
Cd2p = Atom('Cd2+', 48)
In = Atom('In', 49)
In3p = Atom('In3+', 49)
Sn = Atom('Sn', 50)
Sn2p = Atom('Sn2+', 50)
Sn4p = Atom('Sn4+', 50)
Sb = Atom('Sb', 51)
Sb3p = Atom('Sb3+', 51)
Sb5p = Atom('Sb5+', 51)
Te = Atom('Te', 52)
I = Atom('I', 53)
I1m = Atom('I1-', 53)
Xe = Atom('Xe', 54)
Cs = Atom('Cs', 55)
Cs1p = Atom('Cs1+', 55)
Ba = Atom('Ba', 56)
Ba2p = Atom('Ba2+', 56)
La = Atom('La', 57)
La3p = Atom('La3+', 57)
Ce = Atom('Ce', 58)
Ce3p = Atom('Ce3+', 58)
Ce4p = Atom('Ce4+', 58)
Pr = Atom('Pr', 59)
Pr3p = Atom('Pr3+', 59)
Pr4p = Atom('Pr4+', 59)
Nd = Atom('Nd', 60)
Nd3p = Atom('Nd3+', 60)
Pm = Atom('Pm', 61)
Pm3p = Atom('Pm3+', 61)
Sm = Atom('Sm', 62)
Sm3p = Atom('Sm3+', 62)
Eu = Atom('Eu', 63)
Eu2p = Atom('Eu2+', 63)
Eu3p = Atom('Eu3+', 63)
Gd = Atom('Gd', 64)
Gd3p = Atom('Gd3+', 64)
Tb = Atom('Tb', 65)
Tb3p = Atom('Tb3+', 65)
Dy = Atom('Dy', 66)
Dy3p = Atom('Dy3+', 66)
Ho = Atom('Ho', 67)
Ho3p = Atom('Ho3+', 67)
Er = Atom('Er', 68)
Er3p = Atom('Er3+', 68)
Tm = Atom('Tm', 69)
Tm3p = Atom('Tm3+', 69)
Yb = Atom('Yb', 70)
Yb2p = Atom('Yb2+', 70)
Yb3p = Atom('Yb3+', 70)
Lu = Atom('Lu', 71)
Lu3p = Atom('Lu3+', 71)
Hf = Atom('Hf', 72)
Hf4p = Atom('Hf4+', 72)
Ta = Atom('Ta', 73)
Ta5p = Atom('Ta5+', 73)
W = Atom('W', 74)
W6p = Atom('W6+', 74)
Re = Atom('Re', 75)
Os = Atom('Os', 76)
Os4p = Atom('Os4+', 76)
Ir = Atom('Ir', 77)
Ir3p = Atom('Ir3+', 77)
Ir4p = Atom('Ir4+', 77)
Pt = Atom('Pt', 78)
Pt2p = Atom('Pt2+', 78)
Pt4p = Atom('Pt4+', 78)
Au = Atom('Au', 79)
Au1p = Atom('Au1+', 79)
Au3p = Atom('Au3+', 79)
Hg = Atom('Hg', 80)
Hg1p = Atom('Hg1+', 80)
Hg2p = Atom('Hg2+', 80)
Tl = Atom('Tl', 81)
Tl1p = Atom('Tl1+', 81)
Tl3p = Atom('Tl3+', 81)
Pb = Atom('Pb', 82)
Pb2p = Atom('Pb2+', 82)
Pb4p = Atom('Pb4+', 82)
Bi = Atom('Bi', 83)
Bi3p = Atom('Bi3+', 83)
Bi5p = Atom('Bi5+', 83)
Po = Atom('Po', 84)
At = Atom('At', 85)
Rn = Atom('Rn', 86)
Fr = Atom('Fr', 87)
Ra = Atom('Ra', 88)
Ra2p = Atom('Ra2+', 88)
Ac = Atom('Ac', 89)
Ac3p = Atom('Ac3+', 89)
Th = Atom('Th', 90)
Th4p = Atom('Th4+', 90)
Pa = Atom('Pa', 91)
U = Atom('U', 92)
U3p = Atom('U3+', 92)
U4p = Atom('U4+', 92)
U6p = Atom('U6+', 92)
Np = Atom('Np', 93)
Np3p = Atom('Np3+', 93)
Np4p = Atom('Np4+', 93)
Np6p = Atom('Np6+', 93)
Pu = Atom('Pu', 94)
Pu3p = Atom('Pu3+', 94)
Pu4p = Atom('Pu4+', 94)
Pu6p = Atom('Pu6+', 94)
Am = Atom('Am', 95)
Cm = Atom('Cm', 96)
Bk = Atom('Bk', 97)
Cf = Atom('Cf', 98)
Es = Atom("Es", 99)
Fm = Atom("Fm", 100)
Md = Atom("Md", 101)
No = Atom("No", 102)
Lr = Atom("Lr", 103)
Rf = Atom("Rf", 104)
Db = Atom("Db", 105)
Sg = Atom("Sg", 106)
Bh = Atom("Bh", 107)
Hs = Atom("Hs", 108)
Mt = Atom("Mt", 109)
Ds = Atom("Ds", 110)
Rg = Atom("Rg", 111)
Cn = Atom("Cn", 112)
Uut = Atom("Uut", 113)
Uuq = Atom("Uuq", 114)
Uup = Atom("Uup", 115)
Uuh = Atom("Uuh", 116)
Uus = Atom("Uus", 117)
Uuo = Atom("Uuo", 118)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2012-2018 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
miscellaneous functions helpful in the analysis and experiment
"""

import numpy

from .. import config, math


def getangles(peak, sur, inp):
    """
    calculates the chi and phi angles for a given peak

    Parameters
    ----------
    peak :  list or array-like
        hkl for the peak of interest
    sur :   list or array-like
        hkl of the surface
    inp :   list or array-like
        inplane reference peak or direction

    Returns
    -------
    list
        [chi, phi] for the given peak on surface sur with inplane direction inp
        as reference

    Examples
    --------
    To get the angles for the -224 peak on a 111 surface type

    >>> [chi, phi] = getangles([-2, 2, 4], [1, 1, 1], [2, 2, 4])
    """

    # transform input to numpy.arrays
    peak = numpy.array(peak)
    sur = numpy.array(sur)
    inp = numpy.array(inp)

    peak = peak / numpy.linalg.norm(peak)
    sur = sur / numpy.linalg.norm(sur)
    inp = inp / numpy.linalg.norm(inp)

    # calculate reference inplane direction
    inplane = numpy.cross(numpy.cross(sur, inp), sur)
    inplane = inplane / numpy.linalg.norm(inplane)
    if config.VERBOSITY >= config.INFO_ALL:
        print("XU.analyis.getangles: reference inplane direction: ", inplane)

    # calculate inplane direction of peak
    pinp = numpy.cross(numpy.cross(sur, peak), sur)
    pinp = pinp / numpy.linalg.norm(pinp)
    if(numpy.linalg.norm(numpy.cross(sur, peak)) <= config.EPSILON):
        pinp = inplane
    if config.VERBOSITY >= config.INFO_ALL:
        print("XU.analyis.getangles: peaks inplane direction: ", pinp)

    # calculate angles
    r2d = 180. / numpy.pi
    chi = numpy.arccos(numpy.dot(sur, peak)) * r2d
    if(numpy.dot(sur, peak) >= 1. - config.EPSILON):
        chi = 0.
        phi = 0.
    elif(numpy.dot(sur, peak) <= -1. + config.EPSILON):
        chi = 180.
        phi = 0.
    elif(numpy.dot(sur, numpy.cross(inplane, pinp)) <= config.EPSILON):
        if numpy.dot(pinp, inplane) >= 1.0:
            phi = 0.
        elif numpy.dot(pinp, inplane) <= -1.0:
            phi = 180.
        else:
            phi = numpy.sign(numpy.dot(sur, numpy.cross(inplane, pinp))) *\
                numpy.arccos(numpy.dot(pinp, inplane)) * r2d
    else:
        phi = numpy.sign(numpy.dot(sur, numpy.cross(inplane, pinp))) * \
            numpy.arccos(numpy.dot(pinp, inplane)) * r2d
    phi = phi - round(phi / 360.) * 360

    return (chi, phi)


def getunitvector(chi, phi, ndir=(0, 0, 1), idir=(1, 0, 0)):
    """
    return unit vector determined by spherical angles and definition of the
    polar axis and inplane reference direction (phi=0)

    Parameters
    ----------
    chi, phi :  float
        spherical angles (polar and azimuthal) in degree
    ndir :      tuple, list or array-like
        polar/z-axis (determines chi=0)
    idir :      tuple, list or array-like
        azimuthal axis (determines phi=0)
    """
    chi_axis = numpy.cross(ndir, idir)
    v = math.rotarb(ndir, chi_axis, chi)
    v = math.rotarb(v, ndir, phi)
    return v / numpy.linalg.norm(v)


def coplanar_intensity(mat, exp, hkl, thickness, thMono, sample_width=10,
                       beam_width=1):
    """
    Calculates the expected intensity of a Bragg peak from an epitaxial thin
    film measured in coplanar geometry (integration over omega and 2theta in
    angular space!)

    Parameters
    ----------
    mat :           Crystal
        Crystal instance for structure factor calculation
    exp :           Experiment
        Experimental(HXRD) class for the angle calculation
    hkl :           list, tuple or array-like
        Miller indices of the peak to calculate
    thickness :     float
        film thickness in nm
    thMono :        float
        Bragg angle of the monochromator (deg)
    sample_width :  float, optional
        width of the sample along the beam
    beam_width :    float, optional
        width of the beam in the same units as the sample size

    Returns
    -------
    float
        intensity of the peak
    """
    # angle calculation for geometrical factors
    om, chi, phi, tt = exp.Q2Ang(mat.Q(hkl))

    # structure factor calculation
    r = abs(mat.StructureFactor(mat.Q(hkl)))**2

    # polarization factor
    Cmono = numpy.cos(2 * numpy.radians(thMono))
    P = (1 + Cmono * numpy.cos(numpy.radians(tt))**2) / ((1 + Cmono))
    # Lorentz factor to be used when integrating in angular space
    L = 1 / numpy.sin(numpy.radians(tt))
    # shape factor: changing illumination with the incidence angle
    shapef = beam_width / (numpy.sin(numpy.radians(om)) * sample_width)
    if shapef > 1:
        shapef = 1

    # absorption correction
    mu = 1 / (mat.absorption_length() * 1e3)
    mu_eff = mu * (abs(1 / numpy.sin(numpy.radians(om))) +
                   abs(1 / numpy.sin(numpy.radians(tt - om))))
    Nblocks = (1 - numpy.exp(-mu_eff * thickness)) / mu_eff

    return r * P * L * shapef * Nblocks



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2016-2018 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
module providing analytic algebraic functions not implemented in scipy or any
other dependency of xrayutilities. In particular the analytic solution of a
quartic equation which is needed for the solution of the dynamic scattering
equations.
"""

import numpy


def solve_quartic(a4, a3, a2, a1, a0):
    """
    analytic solution [1]_ of the general quartic equation. The solved equation
    takes the form :math:`a_4 z^4 + a_3 z^3 + a_2 z^2 + a_1 z + a_0`

    Returns
    -------
    tuple
        tuple of the four (complex) solutions of aboves equation.

    References
    ----------
    .. [1] http://mathworld.wolfram.com/QuarticEquation.html
    """
    a4 = numpy.asarray(a4)
    a3 = numpy.asarray(a3)
    a2 = numpy.asarray(a2)
    a1 = numpy.asarray(a1)
    a0 = numpy.asarray(a0)

    t01 = 1. / a4
    t02 = a3**2
    t04 = t01 * t01
    t05 = t04 * t01
    t09 = a1 * t01
    t10 = (a3 * t02 * t05) / 8
    t11 = (a3 * a2 * t04) / 2
    t03 = t09 + t10 - t11
    t06 = a2 * t01
    t19 = (3 * t02 * t04) / 8
    t07 = t06 - t19
    t08 = t07**2
    t12 = t03**2
    t13 = a0 * t01
    t14 = t05 * t01
    t15 = t02**2
    t16 = (a2 * t02 * t05) / 16
    t20 = (3 * t14 * t15) / 256
    t21 = (a3 * a1 * t04) / 4
    t17 = t13 + t16 - t20 - t21
    t18 = t17**2
    t22 = t12 / 2.
    t23 = (t07 * t08) / 27
    t24 = 3.**(1./2)
    t25 = t12**2
    t26 = 27 * t25
    t27 = t08**2
    t28 = 4 * t07 * t08 * t12
    t29 = 128 * t08 * t18
    t35 = 256 * t17 * t18
    t36 = 16 * t17 * t27
    t37 = 144 * t07 * t12 * t17
    t30 = t26 + t28 + t29 - t35 - t36 - t37
    t31 = t30.astype(numpy.complex)**(1./2)
    t32 = (t24 * t31) / 18
    t34 = (4 * t07 * t17) / 3
    t33 = t22 + t23 + t32 - t34
    t38 = 1 / t33.astype(numpy.complex)**(1./6)
    t39 = t33**(1./3)
    t40 = 12 * a0 * t01
    t41 = t33**(2./3)
    t42 = 9 * t41
    t43 = (3 * a2 * t02 * t05) / 4
    t46 = 6 * t07 * t39
    t47 = (9 * t14 * t15) / 64
    t48 = 3 * a3 * a1 * t04
    t44 = t08 + t40 + t42 + t43 - t46 - t47 - t48
    t45 = t44.astype(numpy.complex)**(1./2)
    t49 = 6.**(1./2)
    t50 = 27 * t12
    t51 = 2 * t07 * t08
    t52 = 3 * t24 * t31
    t62 = 72 * t07 * t17
    t53 = t50 + t51 + t52 - t62
    t54 = t53.astype(numpy.complex)**(1./2)
    t55 = 3 * t03 * t49 * t54
    t59 = t08 * t45
    t60 = 12 * t17 * t45
    t61 = 9 * t41 * t45
    t63 = 12 * t07 * t39 * t45
    t56 = t55 - t59 - t60 - t61 - t63
    t57 = t56.astype(numpy.complex)**(1./2)
    t58 = 1. / t44.astype(numpy.complex)**(1./4)
    t64 = (t38 * t45) / 6
    t65 = - t55 - t59 - t60 - t61 - t63
    t66 = t65.astype(numpy.complex)**(1./2)

    return (-(a3 * t01) / 4 - (t38 * t45) / 6 - (t38 * t57 * t58) / 6,
            (t38 * t57 * t58) / 6 - (t38 * t45) / 6 - (a3 * t01) / 4,
            t64 - (a3 * t01) / 4 - (t38 * t58 * t66) / 6,
            t64 - (a3 * t01) / 4 + (t38 * t58 * t66) / 6)



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2010-2016 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
module to parse xrayutilities user-specific config file
the parsed values are provide as global constants for the use
in other parts of xrayutilities. The config file with the default constants
is found in the python installation path of xrayutilities. It is however not
recommended to change things there, instead the user-specific config file
~/.xrayutilities.conf or the local xrayutilities.conf file should be used.
"""

import configparser
import os.path
from ast import literal_eval

from . import __path__, utilities_noconf

# so far parsed config variables are
#
# wavelength
# energy
# verbosity
# nthreads
# dynlow
# dynhigh
# epsilon
# database filename for atomic structure factors
# kappa_plane and kappa_angle

xuParser = configparser.ConfigParser()
xuParser.optionxform = str


def trytomake(obj, key, typefunc):
    try:
        obj[key] = typefunc(obj[key])
    except KeyError:
        pass


# read global default values for configuration variables
with open(os.path.join(__path__[0], "xrayutilities_default.conf")) as conffile:
    xuParser.read_file(conffile)

# read user configuration and local configuration if available
cfiles = xuParser.read([
    os.path.expanduser(os.path.join("~", ".xrayutilities.conf")),
    "xrayutilities.conf"])

# set global variables according to configuration
sect = "xrayutilities"
INFO_LOW = xuParser.getint(sect, "info_low")
INFO_ALL = xuParser.getint(sect, "info_all")
DEBUG = xuParser.getint(sect, "debug")

VERBOSITY = xuParser.getint(sect, "verbosity")
try:
    WAVELENGTH = xuParser.getfloat(sect, "wavelength")
except ValueError:
    WAVELENGTH = xuParser.get(sect, "wavelength")
except configparser.NoOptionError:
    pass

try:
    ENERGY = xuParser.getfloat(sect, "energy")
except ValueError:
    ENERGY = xuParser.get(sect, "energy")
except configparser.NoOptionError:
    ENERGY = utilities_noconf.lam2en(utilities_noconf.wavelength(WAVELENGTH))
WAVELENGTH = utilities_noconf.en2lam(utilities_noconf.energy(ENERGY))

# number of threads in parallel section of c-code
NTHREADS = xuParser.getint(sect, "nthreads")

# default parameters for the maplog function
DYNLOW = xuParser.getfloat(sect, "dynlow")
DYNHIGH = xuParser.getfloat(sect, "dynhigh")

# small number needed for error checks
EPSILON = xuParser.getfloat(sect, "epsilon")

# name of the database with atomic scattering factors
DBNAME = xuParser.get(sect, "dbname")

# kappa goniometer specific config parameters
KAPPA_PLANE = xuParser.get(sect, "kappa_plane")
KAPPA_ANGLE = xuParser.getfloat(sect, "kappa_angle")

# parser Powder profile related variables
POWDER = dict()

subsec = 'classoptions'
POWDER[subsec] = dict(xuParser.items("powder"))
trytomake(POWDER[subsec], 'oversampling', int)
for k in ('gaussian_smoother_bins_sigma', 'window_width'):
    trytomake(POWDER[subsec], k, float)

subsec = 'global'
POWDER[subsec] = dict(xuParser.items("powder.global"))
for k in ('diffractometer_radius', 'equatorial_divergence_deg'):
    trytomake(POWDER[subsec], k, float)

subsec = 'emission'
POWDER[subsec] = dict(xuParser.items("powder.emission"))
for k in ('crystallite_size_gauss', 'crystallite_size_lor',
          'strain_lor', 'strain_gauss'):
    trytomake(POWDER[subsec], k, float)
for k in ('emiss_wavelengths', 'emiss_intensities',
          'emiss_gauss_widths', 'emiss_lor_widths'):
    trytomake(POWDER[subsec], k, literal_eval)
if 'emiss_wavelengths' in POWDER[subsec]:
    POWDER[subsec]['emiss_wavelengths'] = tuple(
        utilities_noconf.wavelength(wl) * 1e-10
        for wl in POWDER[subsec]['emiss_wavelengths'])

subsec = 'axial'
POWDER[subsec] = dict(xuParser.items("powder.axial"))
trytomake(POWDER[subsec], 'n_integral_points', int)
for k in ('slit_length_source', 'slit_length_target', 'length_sample',
          'angI_deg', 'angD_deg'):
    trytomake(POWDER[subsec], k, float)

subsec = 'absorption'
POWDER[subsec] = dict(xuParser.items("powder.absorption"))
for k in ('absorption_coefficient', 'sample_thickness'):
    trytomake(POWDER[subsec], k, float)

subsec = 'si_psd'
POWDER[subsec] = dict(xuParser.items("powder.si_psd"))
trytomake(POWDER[subsec], 'si_psd_window_bounds', literal_eval)

subsec = 'receiver_slit'
POWDER[subsec] = dict(xuParser.items("powder.receiver_slit"))
trytomake(POWDER[subsec], 'slit_width', float)

subsec = 'tube_tails'
POWDER[subsec] = dict(xuParser.items("powder.tube_tails"))
for k in ('main_width', 'tail_left', 'tail_right', 'tail_intens'):
    trytomake(POWDER[subsec], k, float)

if VERBOSITY >= DEBUG:
    print("XU.config: xrayutilities configuration files: %s" % repr(cfiles))
    print("xrayutilities configuration:")
    for (name, value) in xuParser.items("xrayutilities"):
        print("%s: %s" % (name, value))
    print("---")



# This file is part of xrayutilities.
#
# xrayutilities is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#
# Copyright (C) 2011-2019 Dominik Kriegner <dominik.kriegner@gmail.com>

"""
xrayutilities.analysis is a package for assisting with the analysis of
x-ray diffraction data, mainly reciprocal space maps

Routines for obtaining line cuts from gridded reciprocal space maps are
offered, with the ability to integrate the intensity perpendicular to the
line cut direction.
"""

from .line_cuts import (get_arbitrary_line, get_omega_scan, get_qx_scan,
                        get_qy_scan, get_qz_scan, get_radial_scan,
                        get_ttheta_scan)
from .misc import coplanar_intensity, getangles, getunitvector
from .sample_align import (area_detector_calib, area_detector_calib_hkl,
                           fit_bragg_peak, linear_detector_calib, miscut_calc,
                           psd_chdeg, psd_refl_align)
